import datetime
import os
import sys
import unittest
from time import sleep
from typing import Type
from unittest import skipIf
from uuid import uuid4

from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.contrib.paxos.application import PaxosApplication, PaxosSystem
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import (
    assert_event_handlers_empty,
    clear_event_handlers,
)
from eventsourcing.system.multiprocess import MultiprocessRunner
from eventsourcing.system.ray import RayRunner
from eventsourcing.system.runner import MultiThreadedRunner
from eventsourcing.tests.base import notquick
from eventsourcing.tests.system_test_fixtures import set_db_uri


class TestPaxosSystem(unittest.TestCase):
    # Use the same system object in all tests.
    system = PaxosSystem(setup_tables=True)

    # Use SQLAlchemy infrastructure (override in subclasses).
    infrastructure_class = SQLAlchemyApplication
    # infrastructure_class = PopoApplication

    def test_single_threaded(self):

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        concrete_system = self.system.bind(self.infrastructure_class)
        with concrete_system as runner:
            paxosapplication0 = runner.processes["paxosapplication0"]
            paxosapplication1 = runner.processes["paxosapplication1"]
            paxosapplication2 = runner.processes["paxosapplication2"]

            started1 = datetime.datetime.now()
            assert isinstance(paxosapplication0, PaxosApplication)
            paxosapplication0.propose_value(key1, value1)
            ended1 = (datetime.datetime.now() - started1).total_seconds()

            # Check each process has expected final value.
            self.assert_final_value(paxosapplication0, key1, value1)
            self.assert_final_value(paxosapplication1, key1, value1)
            self.assert_final_value(paxosapplication2, key1, value1)
            print("Resolved paxos 1 with single thread in %ss" % ended1)

            started2 = datetime.datetime.now()
            paxosapplication1.propose_value(key2, value2)
            ended2 = (datetime.datetime.now() - started2).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosapplication0, key2, value2)
            self.assert_final_value(paxosapplication1, key2, value2)
            self.assert_final_value(paxosapplication2, key2, value2)
            print("Resolved paxos 2 with single thread in %ss" % ended2)

            started3 = datetime.datetime.now()
            paxosapplication2.propose_value(key3, value3)
            ended3 = (datetime.datetime.now() - started3).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosapplication0, key3, value3)
            self.assert_final_value(paxosapplication1, key3, value3)
            self.assert_final_value(paxosapplication2, key3, value3)
            print("Resolved paxos 3 with single thread in %ss" % ended3)

    def test_multi_threaded(self):

        if "TRAVIS_PYTHON_VERSION" in os.environ:
            if self.infrastructure_class is SQLAlchemyApplication:
                self.skipTest(
                    "There's an intermittent problem with the multi-threaded"
                    "runner with SQLAlchemy and Python 3.7 on Travis. Fix me :)."
                )
        set_db_uri()

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        runner = MultiThreadedRunner(
            system=self.system, infrastructure_class=self.infrastructure_class
        )

        with runner:
            paxosapplication0 = runner.processes["paxosapplication0"]
            paxosapplication1 = runner.processes["paxosapplication1"]
            paxosapplication2 = runner.processes["paxosapplication2"]

            started1 = datetime.datetime.now()
            assert isinstance(paxosapplication0, PaxosApplication)
            paxosapplication0.propose_value(key1, value1)
            ended1 = (datetime.datetime.now() - started1).total_seconds()

            # Check each process has expected final value.
            self.assert_final_value(paxosapplication0, key1, value1)
            self.assert_final_value(paxosapplication1, key1, value1)
            self.assert_final_value(paxosapplication2, key1, value1)
            print("Resolved paxos 1 with multi threads in %ss" % ended1)

            started2 = datetime.datetime.now()
            paxosapplication1.propose_value(key2, value2)
            ended2 = (datetime.datetime.now() - started2).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosapplication0, key2, value2)
            self.assert_final_value(paxosapplication1, key2, value2)
            self.assert_final_value(paxosapplication2, key2, value2)
            print("Resolved paxos 2 with multi threads in %ss" % ended2)

            started3 = datetime.datetime.now()
            paxosapplication2.propose_value(key3, value3)
            ended3 = (datetime.datetime.now() - started3).total_seconds()

            # Check each process has a resolution.
            self.assert_final_value(paxosapplication0, key3, value3)
            self.assert_final_value(paxosapplication1, key3, value3)
            self.assert_final_value(paxosapplication2, key3, value3)
            print("Resolved paxos 3 with multi threads in %ss" % ended3)

    def test_multiprocessing(self):
        set_db_uri()

        key1, key2, key3 = uuid4(), uuid4(), uuid4()
        value1, value2, value3 = 11111, 22222, 33333

        self.close_connections_before_forking()

        pipeline_ids = [1, 2, 3]
        runner = MultiprocessRunner(
            system=self.system,
            pipeline_ids=pipeline_ids,
            infrastructure_class=self.infrastructure_class,
        )

        # Start running operating system processes.
        with runner:
            # Get local application object.
            paxosapplication0 = runner.get(
                runner.system.process_classes["paxosapplication0"]
            )
            assert isinstance(paxosapplication0, PaxosApplication)

            # Start proposing values on the different system pipelines.
            paxosapplication0.change_pipeline(1)
            started1 = datetime.datetime.now()
            paxosapplication0.propose_value(key1, value1)

            paxosapplication0.change_pipeline(2)
            started2 = datetime.datetime.now()
            paxosapplication0.propose_value(key2, value2)

            paxosapplication0.change_pipeline(3)
            started3 = datetime.datetime.now()
            paxosapplication0.propose_value(key3, value3)

            # Check all the process applications have expected final values.
            paxosapplication1 = runner.get(
                runner.system.process_classes["paxosapplication1"]
            )
            paxosapplication2 = runner.get(
                runner.system.process_classes["paxosapplication2"]
            )

            assert isinstance(paxosapplication1, PaxosApplication)
            paxosapplication0.repository.use_cache = False
            paxosapplication1.repository.use_cache = False
            paxosapplication2.repository.use_cache = False

            self.assert_final_value(paxosapplication0, key1, value1)
            self.assert_final_value(paxosapplication1, key1, value1)
            self.assert_final_value(paxosapplication2, key1, value1)
            duration1 = (datetime.datetime.now() - started1).total_seconds()
            print("Resolved paxos 1 with multiprocessing in %ss" % duration1)

            self.assert_final_value(paxosapplication0, key2, value2)
            self.assert_final_value(paxosapplication1, key2, value2)
            self.assert_final_value(paxosapplication2, key2, value2)
            duration2 = (datetime.datetime.now() - started2).total_seconds()
            print("Resolved paxos 2 with multiprocessing in %ss" % duration2)

            self.assert_final_value(paxosapplication0, key3, value3)
            self.assert_final_value(paxosapplication1, key3, value3)
            self.assert_final_value(paxosapplication2, key3, value3)
            duration3 = (datetime.datetime.now() - started3).total_seconds()
            print("Resolved paxos 3 with multiprocessing in %ss" % duration3)

    @notquick
    @skipIf(
        sys.version_info[:2] == (3, 6),
        "RayRunner not working with Python36 (pickle issue)",
    )
    def test_ray_performance(self):

        set_db_uri()

        num_pipelines = 2
        pipeline_ids = range(num_pipelines)

        runner = RayRunner(
            system=self.system,
            pipeline_ids=pipeline_ids,
            infrastructure_class=self.infrastructure_class,
            setup_tables=True,
        )
        app_class: Type[PaxosApplication] = runner.get_class("paxosapplication0")

        num_proposals_per_pipeline = 25
        num_proposals = num_pipelines * num_proposals_per_pipeline

        # Propose values.
        proposals = list(((uuid4(), i) for i in range(num_proposals)))

        with runner:

            @retry((KeyError, AssertionError), max_attempts=100, wait=0.25, stall=0)
            def assert_final_value(pipeline_id, key, expected):
                actual = runner.get(app_class, pipeline_id).get_final_value(key)
                self.assertEqual(actual, expected)

            # Start timing (just for fun).
            started = datetime.datetime.now()

            for key, value in proposals:
                pipeline_id = value % len(pipeline_ids)
                print("Proposing key {} value {}".format(key, value))
                runner.get(app_class, pipeline_id).propose_value(key, str(value))
                sleep(0.005)

            # Check final values.
            for key, value in proposals:
                print("Asserting final value for key {} value {}".format(key, value))
                pipeline_id = value % len(pipeline_ids)
                assert_final_value(pipeline_id, key, str(value))

            # Print timing information (just for fun).
            duration = (datetime.datetime.now() - started).total_seconds()
            print(
                "Resolved {} paxoses with ray in {:.4f}s "
                "({:.1f} values/s, {:.4f}s each)".format(
                    num_proposals,
                    duration,
                    num_proposals / duration,
                    duration / num_proposals,
                )
            )

    @notquick
    def test_multiprocessing_performance(self):

        set_db_uri()

        num_pipelines = 2
        pipeline_ids = range(num_pipelines)

        runner = MultiprocessRunner(
            system=self.system,
            pipeline_ids=pipeline_ids,
            infrastructure_class=self.infrastructure_class,
            setup_tables=True,
        )

        num_proposals = 50

        self.close_connections_before_forking()

        with runner:
            sleep(1)

            # Construct an application instance in this process.
            paxosapplication0 = runner.get(
                runner.system.process_classes["paxosapplication0"]
            )

            assert isinstance(paxosapplication0, PaxosApplication)

            # Don't use the cache, so as to keep checking actual database.
            paxosapplication0.repository.use_cache = False

            # Start timing (just for fun).
            started = datetime.datetime.now()

            # Propose values.
            proposals = list(((uuid4(), i) for i in range(num_proposals)))
            for key, value in proposals:
                paxosapplication0.change_pipeline((value % len(pipeline_ids)))
                print("Proposing key {} value {}".format(key, value))
                paxosapplication0.propose_value(key, str(value))
                sleep(0.0)

            # Check final values.
            for key, value in proposals:
                print("Asserting final value for key {} value {}".format(key, value))
                self.assert_final_value(paxosapplication0, key, str(value))

            # Print timing information (just for fun).
            duration = (datetime.datetime.now() - started).total_seconds()
            print(
                "Resolved {} paxoses with multiprocessing in {:.4f}s ({:.4f}s "
                "each)".format(num_proposals, duration, duration / num_proposals)
            )

    @retry((KeyError, AssertionError), max_attempts=100, wait=0.5, stall=0)
    def assert_final_value(self, process, id, value):
        self.assertEqual(process.repository[id].final_value, value)

    def close_connections_before_forking(self):
        """Implemented by the DjangoTestCase class."""
        pass

    def setUp(self) -> None:
        assert_event_handlers_empty()

    def tearDown(self):
        try:
            del os.environ["DB_URI"]
        except KeyError:
            pass
        try:
            assert_event_handlers_empty()
        finally:
            clear_event_handlers()



from abc import abstractmethod
from decimal import Decimal
from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, Tuple
from uuid import UUID, uuid1

from eventsourcing.domain.model.versioning import Upcastable
from eventsourcing.exceptions import EventHashError
from eventsourcing.utils import transcoding_v1
from eventsourcing.utils.hashing import hash_object
from eventsourcing.utils.times import decimaltimestamp
from eventsourcing.utils.topic import get_topic
from eventsourcing.utils.transcoding import ObjectJSONEncoder
from eventsourcing.whitehead import ActualOccasion, TEntity, TEvent


class DomainEvent(Upcastable, ActualOccasion, Generic[TEntity]):
    """
    Base class for domain model events.

    Implements methods to make instances read-only,
    comparable for equality in Python, and have
    recognisable representations.Custom
    To make domain events hashable, this class also
    implements a method to create a cryptographic hash
    of the state of the event.
    """

    __json_encoder_v2__ = ObjectJSONEncoder(sort_keys=True)
    __json_encoder_v1__ = transcoding_v1.ObjectJSONEncoder(sort_keys=True)
    __notifiable__ = True

    def __init__(self, **kwargs: Any):
        """
        Initialises event attribute values directly from constructor kwargs.
        """
        super().__init__()
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        """
        Creates a string representing the type and attribute values of the event.

        :rtype: str
        """
        sorted_items = tuple(sorted(self.__dict__.items()))
        args_strings = ("{0}={1!r}".format(*item) for item in sorted_items)
        args_string = ", ".join(args_strings)
        return "{}({})".format(self.__class__.__qualname__, args_string)

    def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
        """
        Updates 'obj' with values from 'self'.

        Calls the 'mutate()' method.

        Can be extended, but subclasses must call super
        and return an object to their caller.

        :param obj: object (normally a domain entity) to be mutated
        :return: mutated object
        """
        if obj is not None:
            self.mutate(obj)
        return obj

    def mutate(self, obj: TEntity) -> None:
        """
        Updates ("mutates") given 'obj'.

        Intended to be overridden by subclasses, as the most
        concise way of coding a default projection of the event
        (for example into the state of a domain entity).

        The advantage of implementing a default projection
        using this method rather than __mutate__ is that you
        don't need to call super or return a value.

        :param obj: domain entity to be mutated
        """

    def __setattr__(self, key: Any, value: Any) -> None:
        """
        Inhibits event attributes from being updated by assignment.
        """
        raise AttributeError("DomainEvent attributes are read-only")

    def __eq__(self, other: object) -> bool:
        """
        Tests for equality of two event objects.

        :rtype: bool
        """
        return isinstance(other, DomainEvent) and self.__hash__() == other.__hash__()

    # Todo: Do we need this in Python 3?
    def __ne__(self, other: object) -> bool:
        """
        Negates the equality test.

        :rtype: bool
        """
        return not (self == other)

    def __hash__(self) -> int:
        """
        Computes a Python integer hash for an event.

        Supports Python equality and inequality comparisons.

        :return: Python integer hash
        :rtype: int
        """
        attrs = self.__dict__.copy()

        # Involve the topic in the hash, so that different types
        # with same attribute values have different hash values.
        attrs["__event_topic__"] = get_topic(type(self))

        # Calculate the cryptographic hash of the event.
        sha256_hash = self.__hash_object_v2__(attrs)

        # Return the Python hash of the cryptographic hash.
        return hash(sha256_hash)

    @classmethod
    def __hash_object_v2__(cls, obj: dict) -> str:
        """
        Calculates SHA-256 hash of JSON encoded 'obj'.

        :param obj: Object to be hashed.
        :return: SHA-256 as hexadecimal string.
        :rtype: str
        """
        return hash_object(cls.__json_encoder_v2__, obj)

    @classmethod
    def __hash_object_v1__(cls, obj: dict) -> str:
        """
        Calculates SHA-256 hash of JSON encoded 'obj'.

        :param obj: Object to be hashed.
        :return: SHA-256 as hexadecimal string.
        :rtype: str
        """
        return hash_object(cls.__json_encoder_v1__, obj)


class EventWithHash(DomainEvent[TEntity]):
    """
    Base class for domain events with a cryptographic event hash.

    Extends DomainEvent by setting a cryptographic event hash
    when the event is originated, and checking the event hash
    whenever its default projection mutates an object.
    """

    def __init__(self, **kwargs: Any):
        super(EventWithHash, self).__init__(**kwargs)

        # Set __event_topic__ to differentiate events of
        # different types with otherwise equal attributes.
        self.__dict__["__event_topic__"] = get_topic(type(self))

        # Set __event_hash__ with a SHA-256 hash of the event.
        hash_method = self.__hash_object_v2__
        self.__dict__["__event_hash_method_name__"] = hash_method.__name__
        self.__dict__["__event_hash__"] = hash_method(self.__dict__)

    @property
    def __event_hash__(self) -> Any:
        """
        Returns SHA-256 hash of the original state of the event.

        :return: SHA-256 as hexadecimal string.
        :rtype: str
        """
        return self.__dict__.get("__event_hash__")

    def __hash__(self) -> int:
        """
        Computes a Python integer hash for an event,
        using its pre-computed event hash.

        Supports Python equality and inequality comparisons only.

        :return: Python integer hash
        :rtype: int
        """
        # Return the Python hash of the cryptographic hash.
        return hash(self.__event_hash__)

    def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
        """
        Updates 'obj' with values from self.

        Can be extended, but subclasses must call super
        method, and return an object.

        :param obj: object to be mutated
        :return: mutated object
        """

        # Check the hash.
        self.__check_hash__()

        # Call super and return value.
        return super().__mutate__(obj)

    def __check_hash__(self) -> None:
        """
        Raises EventHashError, unless self.__event_hash__ can
        be derived from the current state of the event object.
        """
        state = self.__dict__.copy()
        event_hash = state.pop("__event_hash__")
        method_name = state.get("__event_hash_method_name__", "__hash_object_v1__")
        hash_method = getattr(self, method_name)
        if event_hash != hash_method(state):
            raise EventHashError()


class EventWithOriginatorID(DomainEvent[TEntity]):
    """
    For events that have an originator ID.
    """

    def __init__(self, originator_id: UUID, **kwargs: Any):
        super(EventWithOriginatorID, self).__init__(
            originator_id=originator_id, **kwargs
        )

    @property
    def originator_id(self) -> UUID:
        """
        Originator ID is the identity of the object
        that originated this event.

        :return: A UUID representing the identity of the originator.
        :rtype: UUID
        """
        return self.__dict__["originator_id"]


class EventWithTimestamp(DomainEvent[TEntity]):
    """
    For events that have a timestamp value.
    """

    def __init__(self, timestamp: Optional[Decimal] = None, **kwargs: Any):
        kwargs["timestamp"] = timestamp or decimaltimestamp()
        super(EventWithTimestamp, self).__init__(**kwargs)

    @property
    def timestamp(self) -> Decimal:
        """
        A UNIX timestamp as a Decimal object.
        """
        return self.__dict__["timestamp"]


class EventWithOriginatorVersion(DomainEvent[TEntity]):
    """
    For events that have an originator version number.
    """

    def __init__(self, originator_version: int, **kwargs: Any):
        if not isinstance(originator_version, int):
            raise TypeError("Version must be an integer: {}".format(originator_version))
        super(EventWithOriginatorVersion, self).__init__(
            originator_version=originator_version, **kwargs
        )

    @property
    def originator_version(self) -> int:
        """
        Originator version is the version of the object
        that originated this event.

        :return: A integer representing the version of the originator.
        """
        return self.__dict__["originator_version"]


class EventWithTimeuuid(DomainEvent[TEntity]):
    """
    For events that have an UUIDv1 event ID.
    """

    def __init__(self, event_id: Optional[UUID] = None, **kwargs: Any):
        kwargs["event_id"] = event_id or uuid1()
        super(EventWithTimeuuid, self).__init__(**kwargs)

    @property
    def event_id(self) -> UUID:
        return self.__dict__["event_id"]


class CreatedEvent(DomainEvent[TEntity]):
    """
    Happens when something is created.
    """


class AttributeChangedEvent(DomainEvent[TEntity]):
    """
    Happens when the value of an attribute changes.
    """

    @property
    def name(self) -> str:
        return self.__dict__["name"]

    @property
    def value(self) -> Any:
        return self.__dict__["value"]


class DiscardedEvent(DomainEvent[TEntity]):
    """
    Happens when something is discarded.
    """


class LoggedEvent(DomainEvent[TEntity]):
    """
    Happens when something is logged.
    """


Predicate = Callable[[Sequence[TEvent]], bool]
Handler = Callable[[Sequence[TEvent]], None]

_subscriptions: List[Tuple[Optional[Predicate], Handler]] = []


def subscribe(handler: Handler, predicate: Optional[Predicate] = None) -> None:
    """
    Adds 'handler' to list of event handlers
    to be called if 'predicate' is satisfied.

    If predicate is None, the handler will
    be called whenever an event is published.

    :param callable handler: Will be called when an event is published.
    :param callable predicate: Conditions whether the handler will be called.
    """
    if (predicate, handler) not in _subscriptions:
        _subscriptions.append((predicate, handler))


def unsubscribe(handler: Handler, predicate: Optional[Predicate] = None) -> None:
    """
    Removes 'handler' from list of event handlers
    to be called if 'predicate' is satisfied.

    :param callable handler: Previously subscribed handler.
    :param callable predicate: Previously subscribed predicate.
    """
    if (predicate, handler) in _subscriptions:
        _subscriptions.remove((predicate, handler))


def publish(events: Sequence[TEvent]) -> None:
    """
    Published given 'event' by calling subscribed event
    handlers with the given 'event', except those with
    predicates that are not satisfied by the event.

    Handlers are called in the order they are subscribed.

    :param DomainEvent events: Domain event to be published.
    """
    # A cache of conditions means predicates aren't evaluated
    # more than once for each event.
    cache: Dict[Predicate, bool] = {}
    for predicate, handler in _subscriptions[:]:
        if predicate is None:
            handler(events)
        else:
            cached_condition = cache.get(predicate)
            if cached_condition is None:
                condition = predicate(events)
                cache[predicate] = condition
                if condition:
                    handler(events)
            elif cached_condition is True:
                handler(events)
            else:
                pass


class EventHandlersNotEmptyError(Exception):
    pass


def assert_event_handlers_empty() -> None:
    """
    Raises EventHandlersNotEmptyError, unless
    there are no event handlers subscribed.
    """
    if len(_subscriptions):
        msg = "subscriptions still exist: %s" % _subscriptions
        raise EventHandlersNotEmptyError(msg)


def clear_event_handlers() -> None:
    """
    Removes all previously subscribed event handlers.
    """
    _subscriptions.clear()


def create_timesequenced_event_id() -> UUID:
    return uuid1()


class AbstractSnapshot(ActualOccasion):
    @property
    @abstractmethod
    def topic(self) -> str:
        """
        Path to the class of the snapshotted entity.
        """

    @property
    @abstractmethod
    def state(self) -> Dict[str, Any]:
        """
        State of the snapshotted entity.
        """

    @property
    @abstractmethod
    def originator_id(self) -> UUID:
        """
        ID of the snapshotted entity.
        """

    @property
    @abstractmethod
    def originator_version(self) -> int:
        """
        Version of the last event applied to the entity.
        """

    def __mutate__(self, obj: Optional[TEntity]) -> Optional[TEntity]:
        """
        Reconstructs the snapshotted entity.
        """



from unittest import skip

from eventsourcing.tests.datastore_tests import test_axonserver
from eventsourcing.tests.sequenced_item_tests import base


class AxonServerRecordManagerTestCase(
    test_axonserver.AxonDatastoreTestCase, base.WithRecordManagers
):
    """
    Base class for test cases that need record manager with Axon Server.
    """


class TestAxonServerRecordManagerWithIntegerSequences(
    AxonServerRecordManagerTestCase, base.IntegerSequencedRecordTestCase
):
    """
    Test case for integer sequenced record manager with Axon Server.
    """


@skip("This isn't working at the moment")
class TestSimpleIteratorWithAxonServer(
    AxonServerRecordManagerTestCase, base.SequencedItemIteratorTestCase
):
    """
    Test case for simple iterator in record manager with Axon Server.
    """


@skip("This isn't working at the moment")
class TestThreadedIteratorWithAxonServer(
    AxonServerRecordManagerTestCase, base.ThreadedSequencedItemIteratorTestCase
):
    """
    Test case for threaded iterator in record manager with Axon Server.
    """

    use_named_temporary_file = True



import os

import cassandra.cqlengine
from cassandra import ConsistencyLevel, OperationTimedOut
from cassandra.auth import PlainTextAuthProvider
from cassandra.cluster import NoHostAvailable
from cassandra.cqlengine.management import (
    create_keyspace_simple,
    drop_keyspace,
    sync_table,
)

from eventsourcing.domain.model.decorators import retry
from eventsourcing.exceptions import DatasourceSettingsError
from eventsourcing.infrastructure.datastore import AbstractDatastore, DatastoreSettings

DEFAULT_HOSTS = "127.0.0.1"
DEFAULT_PORT = 9042
DEFAULT_PROTOCOL_VERSION = 3
DEFAULT_DEFAULT_KEYSPACE = "eventsourcing"
DEFAULT_CONSISTENCY_LEVEL = "LOCAL_QUORUM"
DEFAULT_REPLICATION_FACTOR = 1


class CassandraSettings(DatastoreSettings):
    HOSTS = [h.strip() for h in os.getenv("CASSANDRA_HOSTS", DEFAULT_HOSTS).split(",")]
    PORT = int(os.getenv("CASSANDRA_PORT", DEFAULT_PORT))
    PROTOCOL_VERSION = int(
        os.getenv("CASSANDRA_PROTOCOL_VERSION", DEFAULT_PROTOCOL_VERSION)
    )
    DEFAULT_KEYSPACE = os.getenv("CASSANDRA_KEYSPACE", DEFAULT_DEFAULT_KEYSPACE)
    CONSISTENCY_LEVEL = os.getenv(
        "CASSANDRA_CONSISTENCY_LEVEL", DEFAULT_CONSISTENCY_LEVEL
    )
    REPLICATION_FACTOR = os.getenv(
        "CASSANDRA_REPLICATION_FACTOR", DEFAULT_REPLICATION_FACTOR
    )

    def __init__(
        self,
        hosts=None,
        port=None,
        protocol_version=None,
        default_keyspace=None,
        consistency=None,
        replication_factor=None,
        username=None,
        password=None,
    ):
        self.hosts = hosts or self.HOSTS
        self.port = port or self.PORT
        self.protocol_version = protocol_version or self.PROTOCOL_VERSION
        self.default_keyspace = default_keyspace or self.DEFAULT_KEYSPACE
        self.consistency = consistency or self.CONSISTENCY_LEVEL
        self.replication_factor = replication_factor or self.REPLICATION_FACTOR
        self.username = username
        self.password = password


class CassandraDatastore(AbstractDatastore):
    def __init__(self, tables, *args, **kwargs):
        super(CassandraDatastore, self).__init__(*args, **kwargs)
        assert isinstance(tables, tuple), tables
        self.tables = tables

    def setup_connection(self):
        assert isinstance(self.settings, CassandraSettings), self.settings

        # Optionally construct an "auth provider" object.
        if self.settings.username and self.settings.password:
            auth_provider = PlainTextAuthProvider(
                username=self.settings.username, password=self.settings.password
            )
        else:
            auth_provider = None

        # Resolve the consistency level to a driver object.
        try:
            consistency_level_name = self.settings.consistency.upper()
            consistency_level = ConsistencyLevel.name_to_value[consistency_level_name]
        except KeyError:
            msg = "Cassandra consistency level name '{}' not found." "".format(
                self.settings.consistency
            )
            raise DatasourceSettingsError(msg)

        # Use the other self.settings directly.
        cassandra.cqlengine.connection.setup(
            hosts=self.settings.hosts,
            consistency=consistency_level,
            default_keyspace=self.settings.default_keyspace,
            port=self.settings.port,
            auth_provider=auth_provider,
            protocol_version=self.settings.protocol_version,
            lazy_connect=True,
            retry_connect=True,
        )

    def close_connection(self):
        if cassandra.cqlengine.connection.session:
            cassandra.cqlengine.connection.session.shutdown()
        if cassandra.cqlengine.connection.cluster:
            cassandra.cqlengine.connection.cluster.shutdown()

    @retry((NoHostAvailable, OperationTimedOut), max_attempts=10, wait=0.5)
    def setup_tables(self):
        # Avoid warnings about this variable not being set.
        os.environ["CQLENG_ALLOW_SCHEMA_MANAGEMENT"] = "1"

        # Attempt to create the keyspace.
        create_keyspace_simple(
            name=self.settings.default_keyspace,
            replication_factor=self.settings.replication_factor,
        )
        for table in self.tables:
            self.setup_table(table)

    def setup_table(self, table) -> None:
        sync_table(table)

    @retry(NoHostAvailable, max_attempts=10, wait=0.5)
    def drop_tables(self):
        drop_keyspace(name=self.settings.default_keyspace)

    def drop_table(self, *_):
        self.drop_tables()

    @retry(NoHostAvailable, max_attempts=10, wait=0.5)
    def truncate_tables(self):
        for table in self.tables:
            remaining_objects = table.objects.all().limit(10)
            while remaining_objects:
                for obj in remaining_objects:
                    obj.delete()
                remaining_objects = table.objects.all().limit(10)



import time
from collections import OrderedDict, defaultdict, deque
from decimal import Decimal
from threading import Event, Lock
from types import FunctionType
from typing import (
    Any,
    Deque,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

from eventsourcing.application.notificationlog import (
    AbstractNotificationLog,
    NotificationLogReader,
)
from eventsourcing.application.simple import (
    ListOfCausalDependencies,
    ProcessEvent,
    Prompt,
    PromptToPull,
    SimpleApplication,
)
from eventsourcing.application.snapshotting import SnapshottingApplication
from eventsourcing.domain.model.aggregate import (
    BaseAggregateRoot,
    TAggregate,
    TAggregateEvent,
)
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import CausalDependencyFailed, ProgrammingError
from eventsourcing.infrastructure.base import RecordManagerWithTracking, TrackingKwargs
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.whitehead import IterableOfEvents

ListOfAggregateEvents = List[TAggregateEvent]


class PromptToQuit(Prompt):
    pass


class WrappedRepository(Generic[TAggregate, TAggregateEvent]):
    """
    Used to wrap an event sourced repository for use in process application
    policy so that use of, and changes to, domain model aggregates can be
    automatically detected and recorded.

    Implements a "dictionary like" interface, so that aggregates can be
    accessed by ID.
    """

    def __init__(
        self, repository: EventSourcedRepository[TAggregate, TAggregateEvent]
    ) -> None:
        self.repository = repository
        self.retrieved_aggregates: Dict[UUID, TAggregate] = {}
        self.causal_dependencies: List[Tuple[UUID, int]] = []
        self.orm_objs_pending_save: List[Any] = []
        self.orm_objs_pending_delete: List[Any] = []

    def __getitem__(self, entity_id: UUID) -> TAggregate:
        try:
            aggregate = self.retrieved_aggregates[entity_id]
        except KeyError:
            aggregate = self.repository.__getitem__(entity_id)
            self.retrieved_aggregates[entity_id] = aggregate
            self.causal_dependencies.append((aggregate.id, aggregate.__version__))
        return aggregate

    def __contains__(self, entity_id: UUID) -> bool:
        return self.repository.__contains__(entity_id)

    def save_orm_obj(self, orm_obj: Any) -> None:
        """
        Includes orm_obj in "process event", so that projections into
        custom ORM objects is as reliable with respect to sudden restarts
        as "normal" domain event processing in a process application.
        """
        self.orm_objs_pending_save.append(orm_obj)

    def delete_orm_obj(self, orm_obj: Any) -> None:
        """
        Includes orm_obj in "process event", so that projections into
        custom ORM objects is as reliable with respect to sudden restarts
        as "normal" domain event processing in a process application.
        """
        self.orm_objs_pending_delete.append(orm_obj)


class ProcessApplication(SimpleApplication[TAggregate, TAggregateEvent]):
    notification_log_reader_class = NotificationLogReader
    apply_policy_to_generated_events = False

    def __init__(
        self,
        name: str = "",
        policy: Optional[FunctionType] = None,
        setup_table: bool = False,
        use_direct_query_if_available: bool = False,
        notification_log_reader_class: Optional[Type[NotificationLogReader]] = None,
        apply_policy_to_generated_events: bool = False,
        **kwargs: Any
    ):
        self.policy_func = policy
        self.readers: OrderedDict[str, NotificationLogReader] = OrderedDict()
        self.is_reader_position_ok: Dict[str, bool] = defaultdict(bool)
        self._notification_generators: Dict[str, Iterator[Dict[str, Any]]] = {}
        self._policy_lock = Lock()
        self.clock_event: Optional[Event] = None
        self.tick_interval: Optional[Union[float, int]] = None
        self.use_direct_query_if_available = use_direct_query_if_available
        self.notification_log_reader_class = (
            notification_log_reader_class or type(self).notification_log_reader_class
        )
        self.apply_policy_to_generated_events = (
            apply_policy_to_generated_events
            or type(self).apply_policy_to_generated_events
        )

        super(ProcessApplication, self).__init__(
            name=name, setup_table=setup_table, **kwargs
        )

        if self._event_store:
            self.notification_topic_key = (
                self._event_store.record_manager.field_names.topic
            )
            self.notification_state_key = (
                self._event_store.record_manager.field_names.state
            )

        # Publish prompts for any domain events that we persist.
        #  - this means when a process application is used as a simple
        #    application, with calls to aggregate __save__() methods,
        #    then the new events that are stored might need to be processed
        #    by another application. So it can help also to publish a
        #    prompt after the events have been stored so that followers
        #    can be immediately nudged to process the events (otherwise they
        #    would only pick up the events next time they happen to run).
        #    The particular way a prompt published here is actually sent to
        #    any followers is the responsibility of a particular system runner.
        if self._persistence_policy:
            subscribe(
                predicate=self._persistence_policy.is_event,
                handler=self.publish_prompt_for_events,
            )

    def close(self) -> None:
        if self._persistence_policy:
            unsubscribe(
                predicate=self._persistence_policy.is_event,
                handler=self.publish_prompt_for_events,
            )
        super(ProcessApplication, self).close()

    def publish_prompt_for_events(self, _: Optional[IterableOfEvents] = None) -> None:
        """
        Publishes prompt for a given event.

        Used to prompt downstream process application when an event
        is published by this application's model, which can happen
        when application command methods, rather than the process policy,
        are called.

        Wraps exceptions with PromptFailed, to avoid application policy exceptions being
        seen directly in other applications when running synchronously in single thread.
        """

        self.publish_prompt()

    def follow(
        self, upstream_application_name: str, notification_log: AbstractNotificationLog
    ) -> None:
        """
        Sets up process application to follow the given notification log of an
        upstream application.

        :param upstream_application_name: Name of the upstream application.
        :param notification_log: Notification log that will be processed.
        """
        if (
            upstream_application_name == self.name
            and self.apply_policy_to_generated_events
        ):
            raise ProgrammingError(
                "Process application not allowed to follow itself because "
                "its 'apply_policy_to_generated_events' attribute is True."
            )

        # Create a reader.
        reader = self.notification_log_reader_class(
            notification_log,
            use_direct_query_if_available=self.use_direct_query_if_available,
        )
        self.readers[upstream_application_name] = reader

    def run(
        self, prompt: Optional[Prompt] = None, advance_by: Optional[int] = None
    ) -> int:
        """
        Pulls event notifications from notification logs being followed by
        this process application, and processes the contained domain events.

        :param prompt: Optional prompt, specifying a particular notification log.
        :param advance_by: Maximum event notifications to process.
        :return: Returns number of events that have been processed.
        """

        if prompt:
            assert isinstance(prompt, PromptToPull)
            upstream_names = [prompt.process_name]
        else:
            upstream_names = list(self.readers.keys())

        notification_count = 0

        for upstream_name in upstream_names:

            # Set reader position, if necessary.
            if not self.is_reader_position_ok[upstream_name]:
                self.del_notification_generator(upstream_name)
                self.set_reader_position_from_tracking_records(upstream_name)
                self.is_reader_position_ok[upstream_name] = True

            try:
                # Process all the new event notifications in the reader.
                while True:
                    with self._policy_lock:
                        # Get notification generator.
                        generator = self.get_notification_generator(
                            upstream_name, advance_by
                        )
                        try:
                            notification = next(generator)
                        except StopIteration:
                            self.del_notification_generator(upstream_name)
                            break

                        # Increment the notification count.
                        notification_count += 1

                        # Check causal dependencies.
                        self.check_causal_dependencies(
                            upstream_name, notification.get("causal_dependencies")
                        )

                        # Get event from notification.
                        event = self.get_event_from_notification(notification)

                        # Wait for the clock, if there is one.
                        if self.clock_event is not None:
                            self.clock_event.wait()

                        # Process domain event.
                        new_events, new_records = self.process_upstream_event(
                            event, notification["id"], upstream_name
                        )

                    self.take_snapshots(new_events)

                    # Publish a prompt if there are new notifications.
                    # Todo: Optionally send events as prompts, saves pulling
                    #  event if it arrives in order.
                    if any([event.__notifiable__ for event in new_events]):
                        self.publish_prompt()
            except Exception as e:
                # Need to invalidate reader position, so it is refreshed.
                self.is_reader_position_ok[upstream_name] = False
                raise

        return notification_count

    def check_causal_dependencies(self, upstream_name, causal_dependencies_json):
        """
        Checks the causal dependencies are satisfied (have already been processed).

        :param upstream_name: Name of the upstream application being processed.
        :param causal_dependencies_json: Pipelines and positions in notification logs.
        :raises CausalDependencyFailed: If causal dependencies are not satisfied.
        """
        # Decode causal dependencies of the domain event notification.
        if causal_dependencies_json:
            causal_dependencies = self.event_store.event_mapper.json_loads(
                causal_dependencies_json
            )
        else:
            causal_dependencies = []

        record_manager = self.event_store.record_manager
        assert isinstance(record_manager, RecordManagerWithTracking)

        # Check causal dependencies are satisfied.
        for causal_dependency in causal_dependencies:
            # Todo: Check causal dependency on system software version?

            # Check causal dependencies on event notifications.
            assert isinstance(causal_dependency, dict)
            pipeline_id = causal_dependency["pipeline_id"]
            notification_id = causal_dependency["notification_id"]

            has_tracking_record = record_manager.has_tracking_record(
                upstream_application_name=upstream_name,
                pipeline_id=pipeline_id,
                notification_id=notification_id,
            )
            if not has_tracking_record:
                # Invalidate reader position.
                self.is_reader_position_ok[upstream_name] = False

                # Raise exception.
                raise CausalDependencyFailed(
                    {
                        "application_name": self.name,
                        "upstream_name": upstream_name,
                        "pipeline_id": pipeline_id,
                        "notification_id": notification_id,
                    }
                )

    def process_upstream_event(
        self, domain_event: TAggregateEvent, notification_id: int, upstream_name: str
    ) -> Tuple[ListOfAggregateEvents, List]:
        """
        Processes given domain event from an upstream notification log.

        Calls the process application policy, and then records a process event,
        hence recording atomically all new domain events created by the call
        to the policy along with any ORM objects that may result.

        :param domain_event: Domain event to be processed.
        :param notification_id: Position in notification log.
        :param upstream_name: Name of upstream application.
        :return: Returns a list of new domain events.
        """
        cycle_started: Optional[float] = None
        if self.tick_interval is not None:
            cycle_started = time.process_time()

        # Call policy with the upstream event.
        (
            domain_events,
            causal_dependencies,
            orm_objs_pending_save,
            orm_objs_pending_delete,
        ) = self.call_policy(domain_event)

        # Todo: Supplement causal dependencies with system software version?
        #  - this may help to inhibit premature processing when upgrading

        # Todo: Optionally add correlation and causation IDs here. Could be done
        #  in user application code, perhaps with a decorator on policy functions
        #  but having it done here would be more convenient.

        # Todo: Optionally add event and system version numbers in events. This
        #  would help with upcasting them. Could be done in user application code,
        #  with either aggregate or event base class, but having it done here would
        #  be more convenient.

        # Record process event.
        try:
            tracking_kwargs = self.construct_tracking_kwargs(
                notification_id, upstream_name
            )
            process_event = ProcessEvent(
                domain_events=domain_events,
                tracking_kwargs=tracking_kwargs,
                causal_dependencies=causal_dependencies,
                orm_objs_pending_save=orm_objs_pending_save,
                orm_objs_pending_delete=orm_objs_pending_delete,
            )
            new_event_records = self.record_process_event(process_event)

            # Todo: Maybe write one tracking record at the end of a run, if
            #  necessary, or only during a period of time when nothing happens?
        except Exception as exc:
            # Need to purge from the cache relevant entities that
            # have evolved their state past what has been recorded,
            # otherwise strange errors (about version mismatches, or
            # when identifying causal dependencies) can arise.
            if self.repository.use_cache:
                originator_ids = set([event.originator_id for event in domain_events])
                with self.repository.cache_lock:
                    for originator_id in originator_ids:
                        self.repository.cache.pop(originator_id, None)
            raise exc
        else:
            if self.tick_interval is not None:
                assert cycle_started
                # Todo: Change this to use the full cycle time
                #  (improve getting notifications first).
                cycle_ended = time.process_time()
                cycle_time = cycle_ended - cycle_started
                cycle_perc = 100 * (cycle_time) / self.tick_interval
                if cycle_perc > 100:
                    msg = "Warning: {} cycle exceeded tick interval by: {:.2f}%".format(
                        self.name, cycle_perc - 100
                    )
                    print(msg)
        return domain_events, new_event_records

    def get_event_from_notification(
        self, notification: Dict[str, Any]
    ) -> TAggregateEvent:
        """
        Extracts the domain event of an event notification.

        :param notification: The event notification.
        :return: A domain event.
        """
        return self.event_store.event_mapper.event_from_topic_and_state(
            topic=notification[self.notification_topic_key],
            state=notification[self.notification_state_key],
        )

    def get_notification_generator(
        self, upstream_name: str, advance_by: Optional[int]
    ) -> Iterator[Dict[str, Any]]:
        # Dict avoids re-entrant calls to run() starting their own generator,
        # so that notifications are only received once. Was needed in
        # single-threaded runner before it was changed to use iteration not
        # recursion. Hence, probably no longer needed - use reader directly.
        try:
            generator = self._notification_generators[upstream_name]
        except KeyError:
            # Todo: Rename as 'iterator'? We use an iterator, doesn't matter
            #  whether or not it is a generator.
            generator = self.read_reader(upstream_name, advance_by)
            self._notification_generators[upstream_name] = generator
        return generator

    def read_reader(
        self, upstream_name: str, advance_by: Optional[int] = None
    ) -> Iterator[Dict[str, Any]]:
        return self.readers[upstream_name].read(advance_by=advance_by)

    def del_notification_generator(self, upstream_name: str) -> None:
        try:
            del self._notification_generators[upstream_name]
        except KeyError:
            pass

    def take_snapshots(self, new_events: Sequence[TAggregateEvent]) -> None:
        pass

    def set_reader_position_from_tracking_records(self, upstream_name: str) -> None:
        reader = self.readers[upstream_name]
        recorded_position = self.get_recorded_position(upstream_name)
        reader.seek(recorded_position)

    def get_recorded_position(self, upstream_name):
        record_manager = self.event_store.record_manager
        assert isinstance(record_manager, RecordManagerWithTracking)
        recorded_position = record_manager.get_max_tracking_record_id(upstream_name)
        return recorded_position

    def call_policy(
        self, domain_event: TAggregateEvent
    ) -> Tuple[ListOfAggregateEvents, ListOfCausalDependencies, List[Any], List[Any]]:
        """
        Calls the process application policy with the given domain event.

        :param domain_event: Domain event that will be given to the policy.

        :return: Returns a list of domain events, and a list of causal dependencies.
        """
        # Get the application policy.
        policy = self.policy_func or self.policy

        # Wrap the actual repository, so we can collect aggregates.
        wrappedrepo = WrappedRepository(self.repository)

        # Initialise a deque for FIFO queue of unprocessed events.
        unprocessed: Deque[TAggregateEvent] = deque()
        unprocessed.append(domain_event)

        # Initialise a list to collect and return all the generated events.
        all_generated: ListOfAggregateEvents[TAggregateEvent] = []

        # Iterate on the FIFO queue of unprocessed events.
        while len(unprocessed):

            # Get the next unprocessed domain event.
            domain_event = unprocessed.popleft()

            # Call the policy with this domain event.
            returned = policy(wrappedrepo, domain_event)

            # Collect aggregates retrieved by the policy.
            touched = list(wrappedrepo.retrieved_aggregates.values())

            if returned is not None:
                # Convert returned item to list, if necessary.
                if isinstance(returned, Sequence):
                    new_aggregates = returned
                else:
                    new_aggregates = [returned]

                for aggregate in new_aggregates:
                    # Put new aggregates in repository cache (avoids replay).
                    if wrappedrepo.repository.use_cache:
                        wrappedrepo.repository.put_entity_in_cache(
                            aggregate.id, aggregate
                        )

                    # Make new aggregates available in subsequent policy calls.
                    if self.apply_policy_to_generated_events:
                        wrappedrepo.retrieved_aggregates[aggregate.id] = aggregate

                # Extend the list of touched aggregates with
                # the new ones returned from the policy.
                touched.extend(new_aggregates)

            # Collect all the new_events events.
            new_events: ListOfAggregateEvents = self.collect_pending_events(touched)

            # Extend the list of new_events events.
            all_generated.extend(new_events)

            # Enqueue the new events for subsequence processing.
            if self.apply_policy_to_generated_events:
                unprocessed.extend(new_events)

        # Translate causal dependencies from version of entity to position in pipeline.
        causal_dependencies: ListOfCausalDependencies = []
        if self.use_causal_dependencies:
            # Todo: Optionally reference causal dependencies in current pipeline
            #  and then support processing notification from a single pipeline in
            #  parallel, according to dependencies.
            highest: Dict[int, int] = defaultdict(int)
            rm = self.event_store.record_manager
            assert isinstance(rm, RecordManagerWithTracking)
            for entity_id, entity_version in wrappedrepo.causal_dependencies:
                pipeline_id, notification_id = rm.get_pipeline_and_notification_id(
                    entity_id, entity_version
                )
                if pipeline_id is not None and pipeline_id != self.pipeline_id:
                    highest[pipeline_id] = max(notification_id, highest[pipeline_id])

            for pipeline_id, notification_id in highest.items():
                causal_dependencies.append(
                    {"pipeline_id": pipeline_id, "notification_id": notification_id}
                )

        return (
            all_generated,
            causal_dependencies,
            wrappedrepo.orm_objs_pending_save,
            wrappedrepo.orm_objs_pending_delete,
        )

    def policy(
        self,
        repository: WrappedRepository[TAggregate, TAggregateEvent],
        event: TAggregateEvent,
    ) -> Optional[Union[TAggregate, Sequence[TAggregate]]]:
        """
        Empty method, can be overridden in subclasses to implement concrete policy.
        """

    def collect_pending_events(
        self, aggregates: Sequence[TAggregate]
    ) -> ListOfAggregateEvents:
        """
        Collects all the pending events from the given sequence of aggregates.

        :param aggregates: Sequence of aggregates.
        :return: Returns a list of domain events.
        """
        pending_events: ListOfAggregateEvents = []
        num_changed_aggregates = 0
        # This doesn't necessarily obtain events in causal order...
        for aggregate in aggregates:
            batch = aggregate.__batch_pending_events__()
            if len(batch):
                num_changed_aggregates += 1
            pending_events += batch

        # ...so sort pending events across all aggregates.
        if num_changed_aggregates > 1:
            # Sort the events by timestamp.
            #  - this method is intended to establish the correct
            #    causal ordering of all these new events across all aggregates. It
            #    should work if all events are timestamped, all their timestamps
            #    are from the same clock, and none have the same value. If this
            #    doesn't work properly, for some reason, it would be possible when
            #    several aggregates publish events that depend on each other that
            #    concatenating pending events taken from each in turn will be incorrect
            #    and could potentially cause processing errors in a downstream process
            #    application that depends on the correct causal ordering of events. In
            #    the worst case, the events will still be placed correctly in the
            #    aggregate sequence, but if timestamps are skewed and so do not
            #    correctly order the events, the events may be out of order in their
            #    notification log. It is expected in normal usage that these events
            #    are created in the same operating system thread, with timestamps
            #    from the same operating system clock, and so the timestamps will
            #    provide the correct order. However, if somehow different events are
            #    timestamped from different clocks, then problems may occur if those
            #    clocks give timestamps that skew the correct causal order.

            def pick_timestamp(event: BaseAggregateRoot.Event) -> Decimal:
                return event.timestamp

            pending_events.sort(key=pick_timestamp)

        return pending_events

    def construct_tracking_kwargs(
        self, notification_id: int, upstream_application_name: str
    ) -> TrackingKwargs:
        return {
            "application_name": self.name,
            "upstream_application_name": upstream_application_name,
            "pipeline_id": self.pipeline_id,
            "notification_id": notification_id,
        }

    def setup_table(self) -> None:
        super(ProcessApplication, self).setup_table()
        if self._datastore is not None:
            record_manager = self.event_store.record_manager
            assert isinstance(record_manager, RecordManagerWithTracking)
            self._datastore.setup_table(record_manager.tracking_record_class)

    def drop_table(self) -> None:
        super(ProcessApplication, self).drop_table()
        if self.datastore is not None:
            record_manager = self.event_store.record_manager
            assert isinstance(record_manager, RecordManagerWithTracking)
            self.datastore.drop_table(record_manager.tracking_record_class)


class ProcessApplicationWithSnapshotting(SnapshottingApplication, ProcessApplication):
    """
    Supplements process applications that will use snapshotting.
    """
    def take_snapshots(self, new_events: Sequence[TAggregateEvent]) -> None:
        """
        Takes snapshot of aggregates, according to the policy.

        :param new_events: Domain events used to detect if a snapshot is to be taken.
        """
        assert self.snapshotting_policy
        for event in new_events:
            if self.snapshotting_policy.condition([event]):
                self.snapshotting_policy.take_snapshot([event])



from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Sequence
from uuid import UUID

from readerwriterlock.rwlock import RWLockFair

from eventsourcing.exceptions import RecordConflictError
from eventsourcing.infrastructure.base import RecordManagerWithTracking, TrackingKwargs


class PopoNotification(object):
    def __init__(
        self,
        notification_id: int,
        originator_id: UUID,
        originator_version: int,
        topic: str,
        state: str,
    ):
        self.notification_id = notification_id
        self.originator_id = originator_id
        self.originator_version = originator_version
        self.topic = topic
        self.state = state


class PopoRecordManager(RecordManagerWithTracking):
    def __init__(self, *args: Any, **kwargs: Any):
        super(PopoRecordManager, self).__init__(*args, **kwargs)
        self._all_sequence_records: Dict[
            Optional[str], Dict[UUID, Dict[int, NamedTuple]]
        ] = {}
        self._all_sequence_max: Dict = {}
        self._all_tracking_records: Dict = {}
        self._all_tracking_max: Dict = {}
        self._all_notification_records: Dict[Optional[str], Dict[int, Any]] = {}
        self._all_notification_max: Dict = {}
        self._rw_lock: RWLockFair = RWLockFair()

    def all_sequence_ids(self) -> List[UUID]:
        ids: List[UUID] = []
        with self._rw_lock.gen_rlock():
            try:
                ids = list(self._all_sequence_records[self.application_name].keys())
            except KeyError:
                pass
        return ids

    def delete_record(self, record: Any) -> None:
        with self._rw_lock.gen_wlock():
            sequence_records = self._get_sequence_records(record.sequence_id)
            position = getattr(record, self.field_names.position)
            sequence_records.pop(position, None)

    def get_max_notification_id(self) -> int:
        with self._rw_lock.gen_rlock():
            max_record_id = self._get_max_record_id()
        return max_record_id

    def _get_max_record_id(self) -> int:
        try:
            max_notification_id = self._all_notification_max[self.application_name]
        except KeyError:
            return 0
        else:
            return max_notification_id

    def _get_notification_records(self) -> Dict[int, Any]:
        notification_records: Dict[int, Any] = {}
        try:
            notification_records = self._all_notification_records[self.application_name]
        except KeyError:
            pass
        return notification_records

    def get_notification_records(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> Iterable:
        notifications = []
        with self._rw_lock.gen_rlock():
            notification_records = self._get_notification_records()
            if start is None:
                i = 1
            else:
                i = start + 1
            while True:
                if stop is not None and i >= stop + 1:
                    break
                try:
                    notification_record = notification_records[i]
                    notification = PopoNotification(
                        notification_id=notification_record["notification_id"],
                        originator_id=notification_record[
                            "sequenced_item"
                        ].originator_id,
                        originator_version=notification_record[
                            "sequenced_item"
                        ].originator_version,
                        topic=notification_record["sequenced_item"].topic,
                        state=notification_record["sequenced_item"].state,
                    )
                    notifications.append(notification)
                except KeyError:
                    break
                else:
                    i += 1

        return notifications

    def get_max_tracking_record_id(self, upstream_application_name: str) -> int:
        max_id = 0
        with self._rw_lock.gen_rlock():
            try:
                app_records = self._all_tracking_records[self.application_name]
                upstream_records = app_records[upstream_application_name]
            except KeyError:
                pass
            else:
                if len(upstream_records):
                    max_id = max(upstream_records)
        return max_id

    def get_record(self, sequence_id: UUID, position: int) -> Any:
        with self._rw_lock.gen_rlock():
            try:
                return self._get_sequence_records(sequence_id)[position]
            except KeyError:
                raise IndexError(self.application_name, sequence_id, position)

    def get_records(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Sequence[Any]:

        start = None
        if gt is not None:
            start = gt + 1
        if gte is not None:
            if start is None:
                start = gte
            else:
                start = max(start, gte)

        end = None
        if lt is not None:
            end = lt
        if lte is not None:
            if end is None:
                end = lte + 1
            else:
                end = min(end, lte + 1)

        selected_records: List = []
        with self._rw_lock.gen_rlock():
            all_sequence_records = self._get_sequence_records(sequence_id)
            if not len(all_sequence_records):
                return []

            if end is None:
                max_position = self._all_sequence_max[self.application_name][
                    sequence_id
                ]
                end = max_position + 1
            if start is None:
                start = min(all_sequence_records.keys())

            for position in range(start, end):
                try:
                    record = all_sequence_records[position]
                except KeyError:
                    pass
                else:
                    selected_records.append(record)

            if not query_ascending:
                selected_records = list(reversed(selected_records))

            if limit is not None:
                selected_records = list(selected_records)[:limit]

            if query_ascending != results_ascending:
                selected_records = list(reversed(selected_records))

        return selected_records

    def _get_sequence_records(self, sequence_id: UUID) -> Dict:
        try:
            return self._all_sequence_records[self.application_name][sequence_id]
        except KeyError:
            return {}

    def has_tracking_record(
        self, upstream_application_name: str, pipeline_id: int, notification_id: int
    ) -> bool:
        raise NotImplementedError()

    def record_items(self, sequenced_items: Iterable[NamedTuple]) -> None:
        records = self.to_records(sequenced_items)
        self.write_records(records=records)

    def write_records(
        self,
        records: Iterable[Any],
        tracking_kwargs: Optional[TrackingKwargs] = None,
        orm_objs_pending_save: Optional[Sequence[Any]] = None,
        orm_objs_pending_delete: Optional[Sequence[Any]] = None,
    ) -> None:
        with self._rw_lock.gen_wlock():
            # Write event and notification records.
            for record in records:
                self._insert_record(record)

            if tracking_kwargs:
                # Write a tracking record.
                upstream_application_name = tracking_kwargs["upstream_application_name"]
                application_name = tracking_kwargs["application_name"]
                notification_id = tracking_kwargs["notification_id"]
                assert application_name == self.application_name, (
                    application_name,
                    self.application_name,
                )
                try:
                    app_tracking_records = self._all_tracking_records[application_name]
                except KeyError:
                    app_tracking_records = {}
                    self._all_tracking_records[
                        self.application_name
                    ] = app_tracking_records
                try:
                    upstream_tracking_records = app_tracking_records[
                        upstream_application_name
                    ]
                except KeyError:
                    upstream_tracking_records = set()
                    app_tracking_records[
                        upstream_application_name
                    ] = upstream_tracking_records

                if notification_id in upstream_tracking_records:
                    raise RecordConflictError(
                        (application_name, upstream_application_name, notification_id)
                    )
                upstream_tracking_records.add(notification_id)

    def _insert_record(self, sequenced_item: NamedTuple) -> None:
        position = getattr(sequenced_item, self.field_names.position)
        if not isinstance(position, int):
            raise NotImplementedError(
                "Popo record manager only supports sequencing with integers, "
                "but position was a {}".format(type(position))
            )

        sequence_id = getattr(sequenced_item, self.field_names.sequence_id)
        try:
            application_records = self._all_sequence_records[self.application_name]
        except KeyError:
            sequence_records: Dict[int, NamedTuple] = {}
            application_records = {sequence_id: sequence_records}
            self._all_sequence_records[self.application_name] = application_records
            self._all_sequence_max[self.application_name] = {}
        else:
            try:
                sequence_records = application_records[sequence_id]
            except KeyError:
                sequence_records = {}
                application_records[sequence_id] = sequence_records

        if position in sequence_records:
            raise RecordConflictError(position, len(sequence_records))

        if self.notification_id_name:
            # Just make sure we aren't making a gap in the sequence.
            if sequence_records:
                max_position = self._all_sequence_max[self.application_name][
                    sequence_id
                ]
                next_position = max_position + 1
            else:
                next_position = 0
            if position != next_position:
                raise AssertionError(
                    "Next position for sequence {} is {}, not {}".format(
                        sequence_id, next_position, position
                    )
                )

        sequence_records[position] = sequenced_item
        self._all_sequence_max[self.application_name][sequence_id] = position

        # Write a notification record.
        if self.notification_id_name:
            try:
                notification_records = self._all_notification_records[
                    self.application_name
                ]
            except KeyError:
                notification_records = {}
                self._all_notification_records[
                    self.application_name
                ] = notification_records

            next_notification_id = (self._get_max_record_id() or 0) + 1
            notification_records[next_notification_id] = {
                "notification_id": next_notification_id,
                "sequenced_item": sequenced_item,
            }
            self._all_notification_max[self.application_name] = next_notification_id

    def to_records(self, sequenced_items: Iterable[NamedTuple]) -> Iterable[Any]:
        return (PopoStoredEventRecord(s) for s in sequenced_items)


class PopoStoredEventRecord(object):
    """
    Encapsulates sequenced item tuple (containing real event object).

    Allows other attributes to be set, such as notification ID.
    """

    def __init__(self, sequenced_item: NamedTuple):
        self.sequenced_item = sequenced_item

    def __getattr__(self, item: str) -> Any:
        return getattr(self.sequenced_item, item)



from collections import deque
from typing import Any, Deque, Generic, List, Sequence, TypeVar

from eventsourcing.domain.model.entity import (
    DomainEntity,
    EntityWithHashchain,
    TDomainEvent,
    TimestampedVersionedEntity,
)

TAggregate = TypeVar("TAggregate", bound="BaseAggregateRoot")
TAggregateEvent = TypeVar("TAggregateEvent", bound="BaseAggregateRoot.Event")


class BaseAggregateRoot(TimestampedVersionedEntity, Generic[TAggregateEvent]):
    """
    Root entity for an aggregate in a domain driven design.
    """

    class Event(TimestampedVersionedEntity.Event[TAggregate]):
        """Supertype for base aggregate root events."""

    class Created(TimestampedVersionedEntity.Created[TAggregate], Event[TAggregate]):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(
        Event[TAggregate], TimestampedVersionedEntity.AttributeChanged[TAggregate]
    ):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(
        Event[TAggregate], TimestampedVersionedEntity.Discarded[TAggregate]
    ):
        """Triggered when an aggregate root is discarded."""

    def __init__(self, **kwargs: Any) -> None:
        super(BaseAggregateRoot, self).__init__(**kwargs)
        self.__pending_events__: Deque[DomainEntity.Event] = deque()

    def __publish__(self, event: Sequence[TDomainEvent]) -> None:
        """
        Defers publishing event(s) to subscribers, by adding
        event to internal collection of pending events.
        """
        self.__pending_events__.extend(event)

    def __save__(self) -> None:
        """
        Publishes all pending events to subscribers.
        """
        batch_of_events = self.__batch_pending_events__()
        if batch_of_events:
            self.__publish_to_subscribers__(batch_of_events)
            # Don't catch exception and put the events back on the queue.
            # Losing them here is consistent with the behaviour of DomainEntity
            # when an event cannot be stored: the event is effectively lost, the
            # state of the entity must be reset, and the operation repeated.
            # In case of an aggregate sequence conflict, developers need to
            # know what has happened since the last save, so can retry only the
            # command(s) that have caused conflict. Best to save once per command,
            # so the command can be retried cleanly. The purpose of the save
            # method is to allow many events from one command to be persisted
            # together. The save command can be used to persist all the events
            # from many commands, but in case of a failed save after several
            # commands have been executed, it is important to know which
            # commands to retry.

    def __batch_pending_events__(self) -> List[DomainEntity.Event]:
        batch_of_events: List[DomainEntity.Event] = []
        try:
            while True:
                batch_of_events.append(self.__pending_events__.popleft())
        except IndexError:
            pass
        return batch_of_events


TAggregateRootWithHashchainedEvents = TypeVar(
    "TAggregateRootWithHashchainedEvents", bound="AggregateRootWithHashchainedEvents"
)


class AggregateRootWithHashchainedEvents(EntityWithHashchain, BaseAggregateRoot):
    """Extends aggregate root base class with hash-chained events."""

    class Event(
        EntityWithHashchain.Event[TAggregateRootWithHashchainedEvents],
        BaseAggregateRoot.Event[TAggregateRootWithHashchainedEvents],
    ):
        """Supertype for aggregate events."""

    class Created(
        EntityWithHashchain.Created[TAggregateRootWithHashchainedEvents],
        BaseAggregateRoot.Created[TAggregateRootWithHashchainedEvents],
        Event[TAggregateRootWithHashchainedEvents],
    ):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(
        Event[TAggregateRootWithHashchainedEvents],
        BaseAggregateRoot.AttributeChanged[TAggregateRootWithHashchainedEvents],
    ):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(
        Event[TAggregateRootWithHashchainedEvents],
        EntityWithHashchain.Discarded[TAggregateRootWithHashchainedEvents],
        BaseAggregateRoot.Discarded[TAggregateRootWithHashchainedEvents],
    ):
        """Triggered when an aggregate root is discarded."""


# For backwards compatibility.
class AggregateRoot(AggregateRootWithHashchainedEvents):
    """Original name for aggregate root base class with hash-chained events."""

    class Event(
        AggregateRootWithHashchainedEvents.Event[TAggregateRootWithHashchainedEvents]
    ):
        """Supertype for aggregate events."""

    class Created(
        Event[TAggregateRootWithHashchainedEvents],
        AggregateRootWithHashchainedEvents.Created[TAggregateRootWithHashchainedEvents],
    ):
        """Triggered when an aggregate root is created."""

    class AttributeChanged(
        Event[TAggregateRootWithHashchainedEvents],
        AggregateRootWithHashchainedEvents.AttributeChanged[
            TAggregateRootWithHashchainedEvents
        ],
    ):
        """Triggered when an aggregate root attribute is changed."""

    class Discarded(
        Event[TAggregateRootWithHashchainedEvents],
        AggregateRootWithHashchainedEvents.Discarded[
            TAggregateRootWithHashchainedEvents
        ],
    ):
        """Triggered when an aggregate root is discarded."""



import importlib
from typing import Any, Dict, Type

from eventsourcing.domain.model.versioning import Upcastable
from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.whitehead import T


def get_topic(domain_class: type) -> str:
    """
    Returns a string describing a class.

    :param domain_class: A class.
    :returns: A string describing the class.
    """
    return (
        domain_class.__module__
        + "#"
        + getattr(domain_class, "__qualname__", domain_class.__name__)
    )


# Todo: Write documentation for this feature (versioning...).

substitutions: Dict[str, str] = {}


def resolve_topic(topic: str) -> Any:
    """
    Resolves topic to the object it references.

    :param topic: A string describing a code object (e.g. an object class).
    :raises TopicResolutionError: If there is no such class.
    :return: Code object that the topic references.
    """
    # Substitute one topic for another, if so defined.
    #  - this allows classes to be moved and renamed
    topic = substitutions.get(topic, topic)

    # Partition topic into module and class names.
    module_name, _, class_name = topic.partition("#")

    # Import the module.
    try:
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))

    # Identify the class.
    try:
        cls = resolve_attr(module, class_name)
    except AttributeError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    return cls


def resolve_attr(obj: Any, path: str) -> Any:
    """
    A recursive version of getattr for navigating dotted paths.

    :param obj: An object for which we want to retrieve a nested attribute.
    :param path: A dot separated string containing zero or more attribute names.
    :raises AttributeError: If there is no such attribute.
    :return: The attribute referred to by the path.
    """
    if not path:
        return obj
    head, _, tail = path.partition(".")
    head_obj = getattr(obj, head)
    return resolve_attr(head_obj, tail)


def reconstruct_object(obj_class: Type[T], obj_state: Dict[str, Any]) -> T:
    """
    Reconstructs object from given class and state.

    :param obj_class: Class of object to be reconstructed.
    :param obj_state: State of object to be reconstructed.
    :return: Reconstructed object.
    """
    if issubclass(obj_class, Upcastable):
        # Upcast the recorded state using the given class.
        obj_state = obj_class.__upcast_state__(obj_state)

    # Reconstruct the object class instance from the state.
    obj = object.__new__(obj_class)
    obj.__dict__.update(obj_state)
    return obj



from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.infrastructure.sequenceditem import (
    SequencedItem,
    SequencedItemFieldNames,
)
from eventsourcing.infrastructure.sequenceditemmapper import AbstractSequencedItemMapper
from eventsourcing.whitehead import TEvent

DEFAULT_PIPELINE_ID = 0

TrackingKwargs = Dict[str, Union[str, int]]


class AbstractRecordManager(ABC):
    has_integrated_snapshots = False
    can_limit_get_records = True
    can_lt_lte_get_records = True
    can_list_sequence_ids = True
    can_delete_records = True

    def __init__(self, **kwargs: Any):
        """
        Initialises record manager.
        """

    @property
    @abstractmethod
    def record_class(self) -> Any:
        """
        Returns record class to be used by the record manager.
        """

    @abstractmethod
    def record_items(self, sequenced_items: Iterable[NamedTuple]) -> None:
        """
        Writes sequenced items into the datastore.
        """

    def record_item(self, sequenced_item: NamedTuple) -> None:
        """
        Writes sequenced item into the datastore.
        """
        self.record_items([sequenced_item])

    @abstractmethod
    def get_item(self, sequence_id: UUID, position: int) -> NamedTuple:
        """
        Gets sequenced item from the datastore.
        """

    @abstractmethod
    def get_items(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Iterator[NamedTuple]:
        """
        Iterates over records in sequence.
        """

    @abstractmethod
    def get_record(self, sequence_id: UUID, position: int) -> Any:
        """
        Gets record at position in sequence.
        """

    @abstractmethod
    def get_records(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Sequence[Any]:
        """
        Returns records for a sequence.
        """

    @abstractmethod
    def all_sequence_ids(self) -> Iterable[UUID]:
        """
        Returns all sequence IDs.
        """

    @abstractmethod
    def delete_record(self, record: Any) -> None:
        """
        Removes permanently given record from the table.
        """


TRecordManager = TypeVar("TRecordManager", bound=AbstractRecordManager)


class BaseRecordManager(AbstractRecordManager):
    def __init__(
        self,
        record_class: type,
        sequenced_item_class: Type[NamedTuple] = SequencedItem,  # type: ignore
        contiguous_record_ids: bool = False,
        application_name: str = "",
        pipeline_id: int = DEFAULT_PIPELINE_ID,
        **kwargs: Any
    ):
        self._record_class = record_class
        self.sequenced_item_class = sequenced_item_class
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)

        if hasattr(self.record_class, "id"):
            self.notification_id_name = "id"
        elif hasattr(self.record_class, "notification_id"):
            self.notification_id_name = "notification_id"
        else:
            self.notification_id_name = ""

        self.contiguous_record_ids = bool(
            contiguous_record_ids and self.notification_id_name
        )
        if hasattr(self.record_class, "application_name"):
            assert application_name, "'application_name' not set when required"
            assert (
                contiguous_record_ids
            ), "'contiguous_record_ids' not set when required"
        self.application_name = application_name
        if hasattr(self.record_class, "pipeline_id"):
            assert hasattr(
                self.record_class, "application_name"
            ), "'application_name' column not defined"
        self._pipeline_id = pipeline_id

    @property
    def record_class(self) -> type:
        return self._record_class

    @property
    def pipeline_id(self) -> int:
        return self._pipeline_id

    @pipeline_id.setter
    def pipeline_id(self, pipeline_id: int) -> None:
        self._pipeline_id = pipeline_id

    def clone(
        self, application_name: str, pipeline_id: int, **kwargs: Any
    ) -> "BaseRecordManager":
        return type(self)(
            record_class=self.record_class,
            contiguous_record_ids=self.contiguous_record_ids,
            sequenced_item_class=self.sequenced_item_class,
            application_name=application_name,
            pipeline_id=pipeline_id,
            **kwargs
        )

    def get_item(self, sequence_id: UUID, position: int) -> NamedTuple:
        """
        Gets sequenced item from the datastore.
        """
        return self.from_record(self.get_record(sequence_id, position))

    def get_items(
        self,
        sequence_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        query_ascending: bool = True,
        results_ascending: bool = True,
    ) -> Iterator[NamedTuple]:
        """
        Returns sequenced item generator.
        """
        records = self.get_records(
            sequence_id=sequence_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            query_ascending=query_ascending,
            results_ascending=results_ascending,
        )
        for item in map(self.from_record, records):
            yield item

    def list_items(self, *args: Any, **kwargs: Any) -> List[NamedTuple]:
        """
        Returns list of sequenced items.
        """
        return list(self.get_items(*args, **kwargs))

    def to_record(self, sequenced_item: NamedTuple) -> object:
        """
        Constructs a record object from given sequenced item object.
        """
        kwargs = self.get_field_kwargs(sequenced_item)
        # Supply application_name, if needed.
        if hasattr(self.record_class, "application_name"):
            kwargs["application_name"] = self.application_name
        # Supply pipeline_id, if needed.
        if hasattr(self.record_class, "pipeline_id"):
            kwargs["pipeline_id"] = self.pipeline_id
        return self.record_class(**kwargs)

    def from_record(self, record: object) -> NamedTuple:
        """
        Constructs and returns a sequenced item object, from given ORM object.
        """
        kwargs = self.get_field_kwargs(record)
        return self.sequenced_item_class(**kwargs)

    def list_sequence_ids(self) -> List[UUID]:
        return list(self.all_sequence_ids())

    def get_field_kwargs(self, item: object) -> Dict[str, Any]:
        return {name: getattr(item, name) for name in self.field_names}

    def raise_sequenced_item_conflict(self, exp=None) -> None:
        msg = "Position already taken in sequence"
        raise RecordConflictError(exp or msg)

    def raise_index_error(self, position: int) -> None:
        raise IndexError("Sequence index out of range: {}".format(position))

    def raise_record_integrity_error(self, e: Exception) -> None:
        raise RecordConflictError(e)

    def raise_operational_error(self, e: Exception) -> None:
        raise OperationalError(e)


class RecordManagerWithNotifications(BaseRecordManager):
    @abstractmethod
    def get_max_notification_id(self) -> int:
        """Return maximum notification ID in pipeline."""

    @abstractmethod
    def get_notification_records(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> Iterable:
        """
        Returns records sequenced by notification ID, from
        application, for pipeline, in given range.

        Args 'start' and 'stop' are positions in a zero-based
        integer sequence.
        """

    def get_notifications(
        self,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        *args: Any,
        **kwargs: Any
    ) -> Iterable:
        for record in self.get_notification_records(
            start=start, stop=stop, *args, **kwargs
        ):
            yield self.create_notification_from_record(record)

    def create_notification_from_record(self, record):
        notification = {"id": getattr(record, self.notification_id_name)}
        for field_name in self.field_names:
            notification[field_name] = getattr(record, field_name)
        if hasattr(record, "causal_dependencies"):
            notification["causal_dependencies"] = record.causal_dependencies
        return notification


class RecordManagerWithTracking(RecordManagerWithNotifications):
    """
    ACID record managers can write tracking records and event records
    in an atomic transaction, needed for atomic processing in process
    applications.
    """

    tracking_record_field_names = [
        "application_name",
        "upstream_application_name",
        "pipeline_id",
        "notification_id",
    ]

    def __init__(
        self, tracking_record_class: Optional[type] = None, *args: Any, **kwargs: Any
    ) -> None:
        super(RecordManagerWithTracking, self).__init__(*args, **kwargs)
        # assert tracking_record_class is not None
        self.tracking_record_class = tracking_record_class

    def clone(
        self, application_name: str, pipeline_id: int, **kwargs: Any
    ) -> "BaseRecordManager":
        return super().clone(
            application_name=application_name,
            pipeline_id=pipeline_id,
            tracking_record_class=self.tracking_record_class,
            **kwargs
        )

    @abstractmethod
    def write_records(
        self,
        records: Iterable[Any],
        tracking_kwargs: Optional[TrackingKwargs] = None,
        orm_objs_pending_save: Optional[Sequence[Any]] = None,
        orm_objs_pending_delete: Optional[Sequence[Any]] = None,
    ) -> None:
        """
        Writes tracking, event and notification records for a process event.
        :param orm_objs_pending_delete:
        :param orm_objs_pending_save:
        """

    def to_records(self, sequenced_items: Iterable[NamedTuple]) -> Iterable[Any]:
        return map(self.to_record, sequenced_items)

    @abstractmethod
    def get_max_tracking_record_id(self, upstream_application_name: str) -> int:
        """Return maximum tracking record ID for notification from upstream
        application in pipeline."""

    @abstractmethod
    def has_tracking_record(
        self, upstream_application_name: str, pipeline_id: int, notification_id: int
    ) -> bool:
        """
        True if tracking record exists for notification from upstream in pipeline.
        """

    def get_pipeline_and_notification_id(
        self, sequence_id: UUID, position: int
    ) -> Tuple:
        """
        Returns pipeline ID and notification ID for
        event at given position in given sequence.
        """
        # Todo: Optimise query by selecting only two columns,
        #  pipeline_id and id (notification ID)?
        record = self.get_record(sequence_id, position)
        notification_id = getattr(record, self.notification_id_name)
        return record.pipeline_id, notification_id


class SQLRecordManager(RecordManagerWithTracking):
    """
    Common aspects of SQL record managers, such as SQLAlchemy and Django record
    managers.
    """

    # Todo: This makes the subclasses harder to read and probably more brittle. So it
    #  might be better to inline this with the subclasses, so that each looks more
    #  like normal Django or SQLAlchemy code.
    # Todo: Also, the record manager test cases don't cover the notification log and
    #  tracking record functionality needed by ProcessApplication, and should so
    #  that other record managers can more easily be developed.

    def __init__(self, *args: Any, **kwargs: Any):
        super(SQLRecordManager, self).__init__(*args, **kwargs)
        self._insert_select_max = None
        self._insert_values = None
        self._insert_tracking_record = None

    def record_items(self, sequenced_items: Iterable[NamedTuple]) -> None:
        # Convert sequenced item(s) to database record(s).
        records = self.to_records(sequenced_items)

        # Write records.
        self.write_records(records)

    @property
    def insert_select_max(self) -> Any:
        """
        SQL statement that inserts records with contiguous IDs,
        by selecting max ID from indexed table records.
        """
        if self._insert_select_max is None:
            if hasattr(self.record_class, "application_name"):
                # Todo: Maybe make it support application_name without pipeline_id?
                assert hasattr(self.record_class, "pipeline_id"), self.record_class
                tmpl = self._insert_select_max_tmpl + self._where_application_name_tmpl
            else:
                tmpl = self._insert_select_max_tmpl
            self._insert_select_max = self._prepare_insert(
                tmpl=tmpl,
                record_class=self.record_class,
                field_names=list(self.field_names),
            )
        return self._insert_select_max

    def _prepare_insert(
        self,
        tmpl: str,
        record_class: type,
        field_names: List[str],
        placeholder_for_id: bool = False,
    ) -> Any:
        """
        With transaction isolation level of "read committed" this should
        generate records with a contiguous sequence of integer IDs, using
        an indexed ID column, the database-side SQL max function, the
        insert-select-from form, and optimistic concurrency control.
        """
        if (
            hasattr(record_class, "application_name")
            and "application_name" not in field_names
        ):
            field_names.append("application_name")
        if hasattr(record_class, "pipeline_id") and "pipeline_id" not in field_names:
            field_names.append("pipeline_id")
        if (
            hasattr(record_class, "causal_dependencies")
            and "causal_dependencies" not in field_names
        ):
            field_names.append("causal_dependencies")
        if self.notification_id_name:
            if placeholder_for_id:
                if self.notification_id_name not in field_names:
                    field_names.append(self.notification_id_name)

        statement = tmpl.format(
            tablename=self.get_record_table_name(record_class),
            columns=", ".join(field_names),
            placeholders=", ".join([self.make_placeholder(f) for f in field_names]),
            notification_id=self.notification_id_name,
        )
        return statement

    @abstractmethod
    def make_placeholder(self, field_name: str) -> str:
        """
        Returns "placeholder" string for late binding of values to query.

        Depends on record manager's adapted database system or adapted ORM.
        """

    _insert_select_max_tmpl = (
        "INSERT INTO {tablename} ({notification_id}, {columns}) "
        "SELECT COALESCE(MAX({tablename}.{notification_id}), 0) + 1, {placeholders} "
        "FROM "
        "{tablename}"
    )

    @property
    @abstractmethod
    def _where_application_name_tmpl(self) -> str:
        """
        Returns template string for "WHERE" clause of SQL statement.
        """

    @property
    def insert_values(self) -> Any:
        """
        SQL statement that inserts records without ID.
        """
        if self._insert_values is None:
            self._insert_values = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.record_class,
                field_names=list(self.field_names),
            )
        return self._insert_values

    @property
    def insert_tracking_record(self) -> Any:
        """
        SQL statement that inserts tracking records.
        """
        if self._insert_tracking_record is None:
            assert self.tracking_record_class is not None
            self._insert_tracking_record = self._prepare_insert(
                tmpl=self._insert_values_tmpl,
                placeholder_for_id=True,
                record_class=self.tracking_record_class,
                field_names=self.tracking_record_field_names,
            )
        return self._insert_tracking_record

    _insert_values_tmpl = (
        "INSERT INTO {tablename} ({columns}) " "VALUES ({placeholders})"
    )

    @abstractmethod
    def get_record_table_name(self, record_class: type) -> str:
        """
        Returns table name - used in raw queries.

        :rtype: str
        """


class AbstractEventStore(ABC, Generic[TEvent, TRecordManager]):
    """
    Abstract base class for event stores. Defines the methods
    expected of an event store by other classes in the library.
    """

    def __init__(
        self, record_manager: TRecordManager, event_mapper: AbstractSequencedItemMapper,
    ):
        """
        Initialises event store object.


        :param record_manager: record manager
        :param event_mapper: sequenced item mapper
        """
        self.record_manager = record_manager
        self.event_mapper = event_mapper

    @abstractmethod
    def store_events(self, events: Iterable[TEvent]) -> None:
        """
        Put domain event in event store for later retrieval.
        """

    @abstractmethod
    def iter_events(
        self,
        originator_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        is_ascending: bool = True,
        page_size: Optional[int] = None,
    ) -> Iterable[TEvent]:
        """
        Returns iterable of domain events for given entity ID.
        """

    def list_events(self, *args: Any, **kwargs: Any) -> List[TEvent]:
        """
        Returns list of domain events for given entity ID.
        """
        return list(self.iter_events(*args, **kwargs))

    @abstractmethod
    def get_event(self, originator_id: UUID, position: int) -> TEvent:
        """
        Returns a single domain event.
        """

    @abstractmethod
    def get_most_recent_event(
        self, originator_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[TEvent]:
        """
        Returns most recent domain event for given entity ID.
        """

    @abstractmethod
    def all_events(self) -> Iterable[TEvent]:
        """
        Returns all domain events in the event store.

        This works by iterating over all sequences,
        so doesn't return events in order. Use a
        Notification Log to project application state.
        """

    def get_domain_events(
        self,
        originator_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        is_ascending: bool = True,
        page_size: Optional[int] = None,
    ) -> Iterable[TEvent]:
        """
        Deprecated. Please use iter_events() instead.

        Gets domain events from the sequence identified by `originator_id`.

        :param originator_id: ID of a sequence of events
        :param gt: get items after this position
        :param gte: get items at or after this position
        :param lt: get items before this position
        :param lte: get items before or at this position
        :param limit: get limited number of items
        :param is_ascending: get items from lowest position
        :param page_size: restrict and repeat database query
        :return: list of domain events
        """
        return self.iter_events(
            originator_id=originator_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            limit=limit,
            is_ascending=is_ascending,
            page_size=page_size,
        )

    @abstractmethod
    def items_from_events(self, events: Iterable[TEvent]) -> Iterable[NamedTuple]:
        """
        Maps domain event to sequenced item namedtuple.

        :param events: An iterable of events.
        """



from functools import reduce

from eventsourcing.tests.example_application_tests.base import WithExampleApplication
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)

#
## Old stuff...

# This tests using all the domain events in the application to project
# something other than an entity. It's not really customization.

# Todo: Support stopping and resuming when iterating over all events.


class TestGetAllEventFromSQLAlchemy(
    SQLAlchemyRecordManagerTestCase, WithExampleApplication
):
    drop_tables = True

    def test(self):
        with self.construct_application() as app:
            # Create three domain entities.
            entity1 = app.create_new_example("a1", "b1")
            entity2 = app.create_new_example("a2", "b2")
            entity3 = app.create_new_example("a3", "b3")

            # Get all the domain events
            es = app.entity_event_store
            domain_events = es.all_events()

            # Project the events into a set of entity IDs.
            def mutate(state, event):
                assert isinstance(state, set)
                state.add(event.originator_id)
                return state

            all_entity_ids = reduce(mutate, domain_events, set())

            # Check we got all the entity IDs.
            self.assertEqual(all_entity_ids, {entity1.id, entity2.id, entity3.id})

            # Todo: With integer sequenced items, to avoid keeping track of
            # a set of IDs, it would be possible to filter on position==0,
            # and either add that to the database query or put that in a
            # generator that yields all IDs of items with position==0 (rather
            # than yielding items that are not in a set, or keeping adding to
            # a set the same ID once for each version).



from uuid import UUID

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.entity import TimeuuidedEntity
from eventsourcing.domain.model.events import DomainEvent, EventWithTimeuuid
from eventsourcing.infrastructure.cassandra.datastore import (
    CassandraDatastore,
    CassandraSettings,
)
from eventsourcing.infrastructure.cassandra.manager import CassandraRecordManager
from eventsourcing.infrastructure.cassandra.records import TimeuuidSequencedRecord
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase
from eventsourcing.tests.datastore_tests.test_cassandra import (
    DEFAULT_KEYSPACE_FOR_TESTING,
)
from eventsourcing.utils.times import decimaltimestamp_from_uuid

# This test has events with TimeUUID value as the 'event ID'. How easy is it to customize
# the infrastructure to support that? We just need to make a model that uses these events,
# define a suitable database table, and configure the other components. It's easy.

# Firstly, define and entity that uses events with TimeUUIDs.


class ExampleEntity(TimeuuidedEntity):
    def __init__(self, **kwargs):
        super(ExampleEntity, self).__init__(**kwargs)
        self._is_finished = False

    class Started(TimeuuidedEntity.Created, EventWithTimeuuid):
        pass

    class Finished(EventWithTimeuuid, TimeuuidedEntity.Discarded):
        pass

    def finish(self):
        self.__trigger_event__(self.Finished)

    @classmethod
    def start(cls):
        return cls.__create__(event_class=ExampleEntity.Started)


# Define a suitable record class.
#  - class TimeuuidSequencedItem has been moved into the library

# Define an application that uses the entity class and the table
class ExampleApplicationWithTimeuuidSequencedItems(object):
    def __init__(self):
        self.event_store = EventStore(
            record_manager=CassandraRecordManager(
                record_class=TimeuuidSequencedRecord, sequenced_item_class=SequencedItem
            ),
            event_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name="originator_id",
                position_attr_name="event_id",
            ),
        )
        self.repository = EventSourcedRepository(event_store=self.event_store)
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store, persist_event_type=DomainEvent
        )

    def start_entity(self):
        return ExampleEntity.start()

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestDomainEventsWithTimeUUIDs(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestDomainEventsWithTimeUUIDs, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.close_connection()
        super(TestDomainEventsWithTimeUUIDs, self).setUp()

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(TimeuuidSequencedRecord,),
        )

    def test(self):
        with ExampleApplicationWithTimeuuidSequencedItems() as app:
            # Create entity.
            entity1 = app.start_entity()
            self.assertIsInstance(entity1.___initial_event_id__, UUID)
            expected_timestamp = decimaltimestamp_from_uuid(
                entity1.___initial_event_id__
            )
            self.assertEqual(entity1.__created_on__, expected_timestamp)
            self.assertTrue(entity1.__last_modified__, expected_timestamp)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)

            retrieved_obj.finish()
            assert retrieved_obj.id not in app.repository



import weakref
from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Optional, Type, TypeVar

from _weakref import ReferenceType
from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.process import ProcessApplication
from eventsourcing.application.simple import ApplicationWithConcreteInfrastructure
from eventsourcing.exceptions import ProgrammingError
from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID

TProcessApplication = TypeVar("TProcessApplication", bound=ProcessApplication)
TSystemRunner = TypeVar("TSystemRunner", bound="AbstractSystemRunner")
TSystem = TypeVar("TSystem", bound="System")


class System(object):
    """
    A system object has a set of pipeline expressions, which involve
    process application classes. A system object can be run using
    a system runner.

    """

    def __init__(self, *pipeline_exprs: Any, **kwargs: Any):
        """
        Initialises a "process network" system object.

        Each pipeline expression of process classes shows directly which process
        following which other process in the system.

        For example, the pipeline expression (A | B | C) shows that B following A,
        and C following B.

        The pipeline expression (A | A) shows that A following A.

        The pipeline expression (A | B | A) shows that B following A, and A following B.

        The pipeline expressions ((A | B | A), (A | C | A)) are equivalent to (A | B
        | A | C | A).

        :param pipeline_exprs: Pipeline expressions involving process application
        classes.
        """
        self.pipelines_exprs = pipeline_exprs
        self.setup_tables = kwargs.get("setup_tables", False)
        self.infrastructure_class = kwargs.get("infrastructure_class", None)
        self.use_direct_query_if_available = kwargs.get(
            "use_direct_query_if_available", False
        )

        self.session = kwargs.get("session", None)
        self.shared_session = None

        self.process_classes: OrderedDict[str, Type[ProcessApplication]] = OrderedDict()
        for pipeline_expr in self.pipelines_exprs:
            for process_class in pipeline_expr:
                process_name = process_class.create_name()
                if process_name not in self.process_classes:
                    self.process_classes[process_name] = process_class

        self.runner: Optional[ReferenceType] = None
        self.is_session_shared = True
        self.shared_session = None

        # Construct pipeline graph.
        self.downstream_names: OrderedDict[str, List[str]] = OrderedDict()
        self.upstream_names: OrderedDict[str, List[str]] = OrderedDict()

        edges = []
        nodes = []
        for pipeline_expr in self.pipelines_exprs:
            previous_name = None
            for process_class in pipeline_expr:
                process_name = process_class.create_name()
                if process_name not in nodes:
                    nodes.append(process_name)
                if previous_name:
                    edges.append((previous_name, process_name))
                previous_name = process_name

        for process_name in nodes:
            self.downstream_names[process_name] = []
            self.upstream_names[process_name] = []

        for upstream_name, downstream_name in edges:
            self.downstream_names[upstream_name].append(downstream_name)
            self.upstream_names[downstream_name].append(upstream_name)

        self.nodes_of_pipeline_spec = nodes
        self.edges_of_pipeline_spec = edges

    def construct_app(
        self,
        process_class: Type[TProcessApplication],
        infrastructure_class: Optional[
            Type[ApplicationWithConcreteInfrastructure]
        ] = None,
        **kwargs: Any,
    ) -> TProcessApplication:
        """
        Constructs process application from given ``process_class``.
        """

        # If process class isn't already an infrastructure class, then
        # subclass the process class with concrete infrastructure.
        if not issubclass(process_class, ApplicationWithConcreteInfrastructure):

            # Default to PopoApplication infrastructure.
            if infrastructure_class is None:
                infrastructure_class = self.infrastructure_class or PopoApplication

            # Assert that we now have an application with concrete infrastructure.
            if not issubclass(
                infrastructure_class, ApplicationWithConcreteInfrastructure
            ):
                raise ProgrammingError(
                    "Given infrastructure_class {} is not subclass of {}"
                    "".format(
                        infrastructure_class, ApplicationWithConcreteInfrastructure
                    )
                )

            # Subclass the process application class with the infrastructure class.
            process_class = process_class.mixin(infrastructure_class)

        assert issubclass(process_class, ApplicationWithConcreteInfrastructure)

        # Set 'session' and 'setup_table' in kwargs.
        kwargs = dict(kwargs)
        if "session" not in kwargs and process_class.is_constructed_with_session:
            kwargs["session"] = self.session or self.shared_session
        if "setup_tables" not in kwargs and self.setup_tables:
            kwargs["setup_table"] = self.setup_tables

        # Construct the process application.
        app = process_class(**kwargs)

        # Catch the session, so it can be shared.
        if self.session is None and self.shared_session is None:
            if process_class.is_constructed_with_session and self.is_session_shared:
                if self.shared_session is None:
                    self.shared_session = app.session

        assert isinstance(app, ProcessApplication), app
        return app

    def __enter__(self) -> "AbstractSystemRunner":
        """
        Supports running a system object directly as a context manager.

        The system is run with the SingleThreadedRunner.
        """
        from eventsourcing.system.runner import SingleThreadedRunner

        if self.runner:
            raise ProgrammingError("System is already running: {}".format(self.runner))

        runner = SingleThreadedRunner(
            system=self,
            use_direct_query_if_available=self.use_direct_query_if_available,
        )
        runner.start()
        self.runner = runner
        return runner

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Supports usage of a system object as a context manager.
        """

        if self.runner:
            runner: Optional[AbstractSystemRunner] = self.runner
            if runner is not None:
                runner.close()
            self.runner = None

    def bind(
        self: TSystem, infrastructure_class: Type[ApplicationWithConcreteInfrastructure]
    ) -> TSystem:
        """
        Constructs a system object that has an infrastructure class
        from a system object constructed without infrastructure class.

        Raises ProgrammingError if already have an infrastructure class.

        :param infrastructure_class:
        :return: System object that has an infrastructure class.
        :rtype: System
        """
        # Check system doesn't already have an infrastructure class.

        if self.infrastructure_class:
            raise ProgrammingError("System already has an infrastructure class")

        # Clone the system object, and set the infrastructure class.
        system = type(self).__new__(type(self))
        system.__dict__.update(dict(deepcopy(self.__dict__)))
        system.__dict__.update(infrastructure_class=infrastructure_class)
        return system


class AbstractSystemRunner(ABC):
    def __init__(
        self,
        system: System,
        infrastructure_class: Optional[
            Type[ApplicationWithConcreteInfrastructure]
        ] = None,
        setup_tables: bool = False,
        use_direct_query_if_available: bool = False,
    ):
        self.system = system
        self.infrastructure_class = (
            infrastructure_class or self.system.infrastructure_class
        )
        self.setup_tables = setup_tables
        self.use_direct_query_if_available = (
            use_direct_query_if_available or system.use_direct_query_if_available
        )
        self.processes: Dict[str, Any] = {}

    def __enter__(self: TSystemRunner) -> TSystemRunner:
        """
        Supports usage of a system runner as a context manager.
        """
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """
        Supports usage of a system runner as a context manager.
        """
        self.close()

    @abstractmethod
    def start(self) -> None:
        """
        Starts running the system.

        Abstract method which must be implemented on concrete descendants.
        """

    def close(self) -> None:
        """
        Closes a running system.
        """
        self.system.shared_session = None
        if self.processes:
            process: ProcessApplication
            for process in self.processes.values():
                process.close()
            self.processes.clear()

    def get_class(self, process_name: str) -> Type[ProcessApplication]:
        return self.system.process_classes[process_name]

    def get(
        self, process_class: Type[TProcessApplication], pipeline_id=DEFAULT_PIPELINE_ID
    ) -> TProcessApplication:
        process_name = process_class.create_name()
        try:
            process = self.processes[process_name]
        except KeyError:
            process = self._construct_app_by_class(process_class)
        return process

    def _construct_app_by_class(
        self, process_class: Type[TProcessApplication]
    ) -> TProcessApplication:
        process = self.system.construct_app(
            process_class=process_class,
            infrastructure_class=self.infrastructure_class,
            setup_table=self.setup_tables or self.system.setup_tables,
            use_direct_query_if_available=self.use_direct_query_if_available,
        )
        self.processes[process.name] = process
        return process



from unittest import skip

from eventsourcing.application.django import DjangoApplication
from eventsourcing.tests.contrib_tests.paxos_tests import test_paxos_system
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)


class TestPaxosSystemWithDjango(DjangoTestCase, test_paxos_system.TestPaxosSystem):
    infrastructure_class = DjangoApplication

    @skip("Ray runner doesn't run with Django infrastructure, yet")
    def test_ray_performance(self):
        pass



from unittest import TestCase

from eventsourcing.domain.model.aggregate import AggregateRoot, BaseAggregateRoot
from eventsourcing.domain.model.entity import EntityWithECC


class MyBaseAggregateRoot(EntityWithECC, BaseAggregateRoot):
    class Event(EntityWithECC.Event, BaseAggregateRoot.Event):
        pass

    class Created(BaseAggregateRoot.Created, Event):
        pass

    class AttributeChanged(Event, BaseAggregateRoot.AttributeChanged):
        pass

    class Discarded(Event, BaseAggregateRoot.Discarded):
        pass


class TestEntityWithECC(TestCase):
    aggregate_class = MyBaseAggregateRoot

    def test(self):
        # Create first aggregate.
        aggregate1 = self.aggregate_class.__create__(application_name="app1")
        # Get first created event.
        events = aggregate1.__batch_pending_events__()
        created_event1: MyBaseAggregateRoot.Event = events[0]
        # Should have unique event ID.
        self.assertEqual(
            created_event1.event_id, "app1:{}:0".format(str(aggregate1.id))
        )
        # Should be correlated with itself.
        self.assertEqual(created_event1.correlation_id, created_event1.event_id)
        # Should be caused by nothing.
        self.assertEqual(created_event1.causation_id, None)

        # Create second aggregate, caused by processing first created event.
        aggregate2 = self.aggregate_class.__create__(
            processed_event=created_event1, application_name="app2"
        )
        # Get second created event.
        events = aggregate2.__batch_pending_events__()
        created_event2: MyBaseAggregateRoot.Event = events[0]
        # Should have unique event ID.
        self.assertEqual(
            created_event2.event_id, "app2:{}:0".format(str(aggregate2.id))
        )
        # Should be correlated with first created event.
        self.assertEqual(created_event2.correlation_id, created_event1.event_id)
        # Should be caused by first created event.
        self.assertEqual(created_event2.causation_id, created_event1.event_id)

        # Create third aggregate, caused by processing second created event.
        aggregate3 = self.aggregate_class.__create__(
            processed_event=created_event2, application_name="app3"
        )
        # Get third created event.
        events = aggregate3.__batch_pending_events__()
        created_event3: MyBaseAggregateRoot.Event = events[0]
        # Should have unique event ID.
        self.assertEqual(
            created_event3.event_id, "app3:{}:0".format(str(aggregate3.id))
        )
        # Should be correlated with first created event.
        self.assertEqual(created_event3.correlation_id, created_event1.event_id)
        # Should be caused by second created event.
        self.assertEqual(created_event3.causation_id, created_event2.event_id)

        # Trigger attribute changed event.
        aggregate3.__change_attribute__(
            name="_a", value=1, application_name="app3", processed_event=created_event3
        )
        # Get third created event.
        events = aggregate3.__batch_pending_events__()
        changed_event: MyBaseAggregateRoot.Event = events[0]
        # Should have unique event ID.
        self.assertEqual(changed_event.event_id, "app3:{}:1".format(str(aggregate3.id)))
        # Should be correlated with first created event.
        self.assertEqual(changed_event.correlation_id, created_event1.event_id)
        # Should be caused by third created event.
        self.assertEqual(changed_event.causation_id, created_event3.event_id)

        # Trigger discarded event.
        aggregate3.__discard__(application_name="app3", processed_event=changed_event)
        # Get third created event.
        events = aggregate3.__batch_pending_events__()
        discarded_event: MyBaseAggregateRoot.Event = events[0]
        # Should have unique event ID.
        self.assertEqual(
            discarded_event.event_id, "app3:{}:2".format(str(aggregate3.id))
        )
        # Should be correlated with first created event.
        self.assertEqual(discarded_event.correlation_id, created_event1.event_id)
        # Should be caused by attribute changed created event.
        self.assertEqual(discarded_event.causation_id, changed_event.event_id)


class MyAggregateRoot(AggregateRoot, EntityWithECC):
    __subclassevents__ = True


class TestMyAggregateRoot(TestEntityWithECC):
    aggregate_class = MyAggregateRoot



import time
from math import floor
from unittest import skip
from uuid import uuid4

from eventsourcing.domain.model.timebucketedlog import start_new_timebucketedlog
from eventsourcing.example.domainmodel import Example, create_new_example
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sqlalchemy.records import (
    IntegerSequencedNoIDRecord,
    IntegerSequencedWithIDRecord,
    TimestampSequencedNoIDRecord,
    TimestampSequencedWithIDRecord,
)
from eventsourcing.infrastructure.timebucketedlog_reader import (
    TimebucketedlogReader,
    get_timebucketedlog_reader,
)
from eventsourcing.tests.base import notquick
from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.example_application_tests.test_example_application_with_encryption import (
    WithEncryption,
)
from eventsourcing.tests.sequenced_item_tests.test_axonserver_record_manager import (
    AxonServerRecordManagerTestCase,
)
from eventsourcing.tests.sequenced_item_tests.test_cassandra_record_manager import (
    WithCassandraRecordManagers,
)
from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import (
    DjangoTestCase,
)
from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import (
    PopoTestCase,
)
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_record_manager import (
    SQLAlchemyRecordManagerTestCase,
)


@notquick
class PerformanceTestCase(base.WithExampleApplication):
    drop_tables = True

    def test_entity_performance(self):
        """
        Reports on the performance of Example entity and repo.

        NB: This test doesn't actually assert anything, so it isn't really a test.
        """

        with self.construct_application() as app:

            # Initialise dict of entities.
            self.entities = {}

            report_name = type(self).__name__[4:]
            print("\n\n{} report:\n".format(report_name))

            repetitions = 10  # 10

            # NB: Use range(1, 5) to test whether we can get more than 10000 items
            # from Cassandra.
            # Setup a number of entities, with different lengths of event history.
            for i in range(0, 4):

                # Initialise table with other entities.
                num_other_entities = i
                filling = []
                for _ in range(num_other_entities):
                    filling.append(create_new_example(a=1, b=2))

                # b = str([uuid4().hex for _ in range(100000)])
                b = 2
                example = create_new_example(a=1, b=b)
                self.entities[i] = example

                # Beat a number of times.
                num_beats = int(floor(10 ** i))
                start_beating = time.time()
                for _ in range(num_beats):
                    # print("Beat example")
                    example.beat_heart()

                    for other in filling:
                        other.beat_heart()

                total_beats = num_beats * (1 + len(filling))
                time_beating = time.time() - start_beating
                try:
                    beats_per_second = total_beats / time_beating
                except ZeroDivisionError as e:
                    print(
                        "Warning: beats per second {} / {}: {}".format(
                            total_beats, time_beating, e
                        )
                    )
                    beats_per_second = -999999999999.99999999
                try:
                    beat_period = time_beating / total_beats
                except ZeroDivisionError as e:
                    print(
                        "Warning: beat period {} / {}: {}".format(
                            time_beating, total_beats, e
                        )
                    )
                    beat_period = -999999999999.9999999

                print(
                    "Time to beat {} times: {:.4f}s ({:.0f} beats/s, {:.6f}s each)"
                    "".format(
                        num_beats,
                        time_beating / (1 + num_other_entities),
                        beats_per_second,
                        beat_period,
                    )
                )

                # Get the last n events from the repo.
                def last_n(n):
                    n = min(n, num_beats + 1)
                    assert isinstance(app.example_repository.event_store, EventStore)
                    start_last_n = time.time()
                    events = app.example_repository.event_store.list_events(
                        originator_id=example.id, gt=num_beats - n,
                    )
                    assert len(events) == n, "Asked for %s but got %s" % (
                        n,
                        len(events),
                    )
                    time_last_n = (time.time() - start_last_n) / repetitions

                    # num_retrieved_events = len(last_n_stored_events)
                    events_per_second = n / time_last_n
                    print(
                        (
                            "Time to get last {:>"
                            + str(i + 1)
                            + "} events after {} events: {:.6f}s ({:.0f} events/s)"
                            ""
                        ).format(n, num_beats + 1, time_last_n, events_per_second)
                    )

                for j in range(0, i + 1):
                    last_n(10 ** j)

                # Get the entity by replaying all events (which it must since there
                # isn't a snapshot).
                start_replay = time.time()
                for _ in range(repetitions):
                    example = app.example_repository[example.id]
                    assert isinstance(example, Example)
                    heartbeats = example.count_heartbeats()
                    assert heartbeats == num_beats, (heartbeats, num_beats)

                time_replaying = (time.time() - start_replay) / repetitions
                print(
                    "Time to replay {} beats: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                    "".format(
                        num_beats,
                        time_replaying,
                        num_beats / time_replaying,
                        time_replaying / num_beats,
                    )
                )

                # Take snapshot, and beat heart a few more times.
                app.example_repository.take_snapshot(example.id)

                extra_beats = 4
                for _ in range(extra_beats):
                    example.beat_heart()
                num_beats += extra_beats

                # Get the entity using snapshot and replaying events since the snapshot.
                start_replay = time.time()
                for _ in range(repetitions):
                    example = app.example_repository[example.id]
                time_replaying = (time.time() - start_replay) / repetitions

                events_per_second = (
                    extra_beats + 1
                ) / time_replaying  # +1 for the snapshot event
                beats_per_second = num_beats / time_replaying
                print(
                    "Time to replay snapshot with {} extra beats: {:.6f}s ({:.0f} "
                    "events/s, {:.0f} beats/s)"
                    "".format(
                        extra_beats, time_replaying, events_per_second, beats_per_second
                    )
                )

                print("")

    def test_log_performance(self):

        with self.construct_application() as app:
            example_id = uuid4()
            log = start_new_timebucketedlog(example_id, bucket_size="year")
            log_reader = get_timebucketedlog_reader(log, app.log_event_store)

            # Write a load of messages.
            start_write = time.time()
            number_of_messages = 111
            events = []
            for i in range(number_of_messages):
                event = log.log_message("Logger message number {}".format(i))
                events.append(event)
            time_to_write = time.time() - start_write
            print(
                "Time to log {} messages: {:.2f}s ({:.0f} messages/s, {:.6f}s each)"
                "".format(
                    number_of_messages,
                    time_to_write,
                    number_of_messages / time_to_write,
                    time_to_write / number_of_messages,
                )
            )

            # Read pages of messages in descending order.
            # - get a limited number until a time, then use the earliest in that list
            # as the position
            position = events[-1].timestamp

            page_size = 10

            # Page back through the log in reverse chronological order.
            previous_position = None
            count_pages = 0
            total_time_to_read = 0
            total_num_reads = 0
            while True:
                start_read = time.time()
                (
                    page_of_events,
                    next_position,
                ) = self.get_message_logged_events_and_next_position(
                    log_reader, position, page_size
                )
                time_to_read = time.time() - start_read
                total_time_to_read += time_to_read
                total_num_reads += 1
                count_pages += 1
                if next_position is None:
                    break
                else:
                    previous_position, position = position, next_position

            # Check we got to the end of the line.
            self.assertEqual(count_pages, 11)
            self.assertIsNone(next_position)
            self.assertTrue(previous_position)

            # Page forward through the log in chronological order.
            count_pages = 0
            position = None
            while True:
                start_read = time.time()
                (
                    page_of_events,
                    next_position,
                ) = self.get_message_logged_events_and_next_position(
                    log_reader, position, page_size, is_ascending=True
                )
                time_to_read = time.time() - start_read
                total_time_to_read += time_to_read
                total_num_reads += 1
                count_pages += 1
                if next_position is None:
                    break
                else:
                    position = next_position

            self.assertEqual(count_pages, 11)
            self.assertIsNone(next_position)
            self.assertTrue(previous_position)

            reads_per_second = total_num_reads / total_time_to_read
            messages_per_second = reads_per_second * number_of_messages
            print(
                "Time to read {} pages of logged messages: {:.6f}s ({:.0f} pages/s, "
                "{:.0f} messages/s))"
                "".format(
                    total_num_reads,
                    total_time_to_read,
                    reads_per_second,
                    messages_per_second,
                )
            )

    def get_message_logged_events_and_next_position(
        self, log_reader, position, page_size, is_ascending=False
    ):
        assert isinstance(log_reader, TimebucketedlogReader), type(log_reader)
        assert isinstance(page_size, int), type(page_size)
        assert isinstance(is_ascending, bool)
        if is_ascending:
            gt = position
            lt = None
        else:
            lt = position
            gt = None

        events = log_reader.get_events(
            gt=gt, lt=lt, limit=page_size + 1, is_ascending=is_ascending
        )
        events = list(events)
        if len(events) == page_size + 1:
            next_position = events.pop().timestamp
        else:
            next_position = None
        return events, next_position


@notquick
class TestCassandraPerformance(WithCassandraRecordManagers, PerformanceTestCase):
    pass


@notquick
class TestDjangoPerformance(DjangoTestCase, PerformanceTestCase):
    pass


@notquick
class TestDjangoPerformanceWithEncryption(WithEncryption, TestDjangoPerformance):
    pass


@notquick
class TestPopoPerformance(PopoTestCase, PerformanceTestCase):
    @skip("Popo record manager only supports sequencing with integers")
    def test_log_performance(self):
        pass


@notquick
class TestCassandraPerformanceWithEncryption(WithEncryption, TestCassandraPerformance):
    pass


@notquick
class TestSQLAlchemyPerformance(SQLAlchemyRecordManagerTestCase, PerformanceTestCase):
    def construct_entity_record_manager(self):
        return self.factory.construct_record_manager(
            record_class=IntegerSequencedWithIDRecord
        )

    def construct_log_record_manager(self):
        return self.factory.construct_record_manager(
            record_class=TimestampSequencedWithIDRecord
        )


@notquick
class TestAxonServerPerformance(AxonServerRecordManagerTestCase, PerformanceTestCase):
    def construct_entity_record_manager(self):
        return self.factory.construct_record_manager(record_class=None)

    def construct_log_record_manager(self):
        return self.factory.construct_record_manager(record_class=None)

    @skip("Axon doesn't support sequencing by timestamp, so don't test log performance")
    def test_log_performance(self):
        pass


@notquick
class TestSQLAlchemyPerformanceNoID(TestSQLAlchemyPerformance):
    def construct_entity_record_manager(self):
        return self.factory.construct_record_manager(
            record_class=IntegerSequencedNoIDRecord
        )

    def construct_log_record_manager(self):
        return self.factory.construct_record_manager(
            record_class=TimestampSequencedNoIDRecord
        )


@notquick
class TestSQLAlchemyPerformanceWithEncryption(
    WithEncryption, TestSQLAlchemyPerformance
):
    pass


# Avoid running abstract test case.
del PerformanceTestCase



import os

from cassandra import InvalidRequest
from cassandra.cqlengine.functions import Token
from cassandra.cqlengine.query import BatchQuery, LWTException

from eventsourcing.exceptions import ProgrammingError
from eventsourcing.infrastructure.base import BaseRecordManager


class CassandraRecordManager(BaseRecordManager):
    def record_items(self, sequenced_items):
        if not isinstance(sequenced_items, list):
            sequenced_items = list(sequenced_items)

        if len(sequenced_items) > 1:
            b = BatchQuery()
            for item in sequenced_items:
                assert isinstance(item, self.sequenced_item_class), (
                    type(item),
                    self.sequenced_item_class,
                )
                kwargs = self.get_field_kwargs(item)
                self.record_class.batch(b).if_not_exists().create(**kwargs)
            try:
                b.execute()
            except LWTException:
                self.raise_sequenced_item_conflict()
        elif len(sequenced_items) == 1:
            record = self.to_record(sequenced_items[0])
            try:
                record.save()
            except LWTException:
                self.raise_sequenced_item_conflict()

    def get_record(self, sequence_id, position):
        kwargs = {
            self.field_names.sequence_id: sequence_id,
            "{}__eq".format(self.field_names.position): position,
        }
        query = self.filter(**kwargs)
        try:
            record = list(query)[0]
        except IndexError:
            self.raise_index_error(position)
        else:
            return record

    def get_records(
        self,
        sequence_id,
        gt=None,
        gte=None,
        lt=None,
        lte=None,
        limit=None,
        query_ascending=True,
        results_ascending=True,
    ):

        assert limit is None or limit >= 1, limit
        assert not (gte and gt)
        assert not (lte and lt)

        kwargs = {self.field_names.sequence_id: sequence_id}
        query = self.filter(**kwargs)

        if query_ascending:
            query = query.order_by(self.field_names.position)

        position_name = self.field_names.position
        if gt is not None:
            kwargs = {"{}__gt".format(position_name): gt}
            query = query.filter(**kwargs)
        if gte is not None:
            kwargs = {"{}__gte".format(position_name): gte}
            query = query.filter(**kwargs)
        if lt is not None:
            kwargs = {"{}__lt".format(position_name): lt}
            query = query.filter(**kwargs)
        if lte is not None:
            kwargs = {"{}__lte".format(position_name): lte}
            query = query.filter(**kwargs)

        if limit is not None:
            query = query.limit(limit)

        items = list(query)

        if results_ascending != query_ascending:
            items.reverse()

        return items

    def all_sequence_ids(self):
        sequence_id_page_size = int(os.getenv("SEQUENCE_ID_PAGE_SIZE") or "1")
        assert sequence_id_page_size > 0, sequence_id_page_size
        query = self.record_class.objects.all().limit(sequence_id_page_size)

        page = list(query)
        # # Resume if possible.
        # if resume is None:
        #     page = list(query)
        # else:
        #     page = list(query.filter(pk__token__gt=Token(resume)))

        while page:
            for record in page:
                yield record.pk
            last = page[-1]
            page = list(query.filter(pk__token__gt=Token(last.pk)))

    def delete_record(self, record):
        assert isinstance(record, self.record_class), type(record)
        try:
            record.delete()
        except InvalidRequest as e:
            raise ProgrammingError(e)

    def filter(self, **kwargs):
        return self.record_class.objects.filter(**kwargs)



from typing import Any, NamedTuple, Optional, Type

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID, AbstractRecordManager
from eventsourcing.infrastructure.datastore import AbstractDatastore
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.datastore import (
    SQLAlchemyDatastore,
    SQLAlchemySettings,
)
from eventsourcing.infrastructure.sqlalchemy.manager import SQLAlchemyRecordManager
from eventsourcing.infrastructure.sqlalchemy.records import (
    IntegerSequencedWithIDRecord,
    NotificationTrackingRecord,
    SnapshotRecord,
    StoredEventRecord,
    TimestampSequencedNoIDRecord,
)
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder


class SQLAlchemyInfrastructureFactory(InfrastructureFactory):
    """
    Infrastructure factory for SQLAlchemy infrastructure.
    """

    record_manager_class = SQLAlchemyRecordManager
    integer_sequenced_record_class = IntegerSequencedWithIDRecord
    timestamp_sequenced_record_class = TimestampSequencedNoIDRecord
    snapshot_record_class = SnapshotRecord
    tracking_record_class = NotificationTrackingRecord

    def __init__(
        self,
        session: Any,
        uri: Optional[str] = None,
        pool_size: Optional[int] = None,
        tracking_record_class: Optional[type] = None,
        *args: Any,
        **kwargs: Any
    ):
        super(SQLAlchemyInfrastructureFactory, self).__init__(*args, **kwargs)
        self.session = session
        self.uri = uri
        self.pool_size = pool_size
        self._tracking_record_class = tracking_record_class

    def construct_integer_sequenced_record_manager(
        self, **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs SQLAlchemy record manager.

        :return: An SQLAlchemy record manager.
        :rtype: SQLAlchemyRecordManager
        """
        tracking_record_class = (
            self._tracking_record_class or self.tracking_record_class
        )
        return super(
            SQLAlchemyInfrastructureFactory, self
        ).construct_integer_sequenced_record_manager(
            tracking_record_class=tracking_record_class, **kwargs
        )

    def construct_record_manager(
        self,
        record_class: Optional[type],
        sequenced_item_class: Optional[Type[NamedTuple]] = None,
        **kwargs: Any
    ) -> AbstractRecordManager:
        """
        Constructs SQLAlchemy record manager.

        :return: An SQLAlchemy record manager.
        :rtype: SQLAlchemyRecordManager
        """
        return super(SQLAlchemyInfrastructureFactory, self).construct_record_manager(
            record_class,
            sequenced_item_class=sequenced_item_class,
            session=self.session,
            **kwargs
        )

    def construct_datastore(self) -> Optional[AbstractDatastore]:
        """
        Constructs SQLAlchemy datastore.

        :rtype: SQLAlchemyDatastore
        """
        datastore = SQLAlchemyDatastore(
            settings=SQLAlchemySettings(uri=self.uri, pool_size=self.pool_size),
            session=self.session,
        )
        if self.session is None:
            assert datastore.session, "Datastore object session is None"
            self.session = datastore.session
        return datastore


def construct_sqlalchemy_eventstore(
    session: Any,
    sequenced_item_class: Optional[Type[NamedTuple]] = None,
    sequence_id_attr_name: Optional[str] = None,
    position_attr_name: Optional[str] = None,
    json_encoder_class: Optional[Type[ObjectJSONEncoder]] = None,
    json_decoder_class: Optional[Type[ObjectJSONDecoder]] = None,
    cipher: Optional[AESCipher] = None,
    record_class: Optional[type] = None,
    contiguous_record_ids: bool = False,
    application_name: Optional[str] = None,
    pipeline_id: int = DEFAULT_PIPELINE_ID,
) -> EventStore:
    sequenced_item_class = sequenced_item_class or StoredEvent  # type: ignore
    sequenced_item_mapper = SequencedItemMapper[DomainEvent](
        sequenced_item_class=sequenced_item_class,
        sequence_id_attr_name=sequence_id_attr_name,
        position_attr_name=position_attr_name,
        json_encoder_class=json_encoder_class,
        json_decoder_class=json_decoder_class,
        cipher=cipher,
    )
    factory = SQLAlchemyInfrastructureFactory(
        session=session,
        integer_sequenced_record_class=record_class or StoredEventRecord,
        sequenced_item_class=sequenced_item_class,
        contiguous_record_ids=contiguous_record_ids,
        application_name=application_name,
        pipeline_id=pipeline_id,
    )
    record_manager = factory.construct_integer_sequenced_record_manager()
    event_store = EventStore[DomainEvent, AbstractRecordManager](
        record_manager=record_manager, event_mapper=sequenced_item_mapper
    )
    return event_store



from typing import Iterable, NamedTuple, TypeVar


class Event(object):
    """
    "I shall use the term 'event' in the more general sense
    of a nexus of actual occasions, inter-related in some
    determinate fashion in one extensive quantum. An actual
    occasion is the limiting type of an event with only one
    member."

    Alfred North Whitehead, 1929
    """


class ActualOccasion(Event):
    """
    "'Actual entities' -- also termed 'actual occasions' -- are the final real things
    of which the world is made up. There is no going behind actual entities
    to find anything more real."

    "Just as 'potentiality for process' is the meaning of the more general term
    'entity' or 'thing'; so 'decision' is the additional meaning imported by the
    word 'actual' into the phrase 'actual entity'. 'Actuality' is the decision
    amid 'potentiality'. It represents stubborn fact which cannot be evaded."

    "Actual entities perish, but do not change; they are what they are."

    Alfred North Whitehead, 1929
    """


class EnduringObject(Event):
    """
    "The notions of 'social order' and of 'personal order' cannot be omitted
    from this preliminary sketch. A 'society' in the sense in which that term
    is here used, is a nexus with social order; and an 'enduring object' or
    'enduring creature' is a society whose social order has taken the special
    form of 'personal order.'"

    "A nexus enjoys 'personal order' when (a) it is a 'society' and when
    the genetic relatedness of its members orders these members 'serially'."

    Alfred North Whitehead, 1929
    """


T = TypeVar("T")

TEvent = TypeVar("TEvent", bound=ActualOccasion)
TEntity = TypeVar("TEntity", bound=EnduringObject)
SEntity = TypeVar("SEntity", bound=EnduringObject)

IterableOfEvents = Iterable[ActualOccasion]
IterableOfItems = Iterable[NamedTuple]



from abc import ABCMeta
from typing import Iterator, Union

# Need to deal with the fact that Python3.6 had GenericMeta.
try:
    from typing import GenericMeta

    ABCMeta = GenericMeta  # type: ignore

except ImportError:
    pass
# Todo: Delete above try/except when dropping support for Python 3.6.


PipeableExpression = Union["PipelineExpression", "PipeableMetaclass"]


class PipelineExpression(object):
    """
    Implements a left-to-right association between two objects.
    """

    def __init__(self, left: PipeableExpression, right: PipeableExpression):
        self.left = left
        self.right = right

    def __or__(self, other: PipeableExpression) -> PipeableExpression:
        return PipelineExpression(self, other)

    def __iter__(self) -> Iterator["PipeableMetaclass"]:
        for term in (self.left, self.right):
            if isinstance(term, PipelineExpression):
                for item in term:
                    yield item
            else:
                assert isinstance(term, PipeableMetaclass), term
                yield term


class PipeableMetaclass(ABCMeta):
    """
    Meta class for pipeable classes.
    """

    def __or__(self, other: PipeableExpression) -> PipelineExpression:
        """
        Implements bitwise or operator '|' as a pipe between pipeable classes.
        """
        return PipelineExpression(self, other)

    def __iter__(self) -> Iterator["PipeableMetaclass"]:
        yield self


class Pipeable(metaclass=PipeableMetaclass):
    """
    Base class for pipeable classes.
    """



from eventsourcing.tests.example_application_tests import base
from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import (
    PopoTestCase,
)


class TestExampleApplicationWithPopo(PopoTestCase, base.ExampleApplicationTestCase):
    def construct_log_record_manager(self):
        return self.construct_entity_record_manager()

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()



# import datetime
#
# from eventsourcing.example.application import ApplicationWithPersistencePolicies
#
#
# class SuffixTreeApplication(ApplicationWithPersistencePolicies):
#     def __init__(self, **kwargs):
#         super(SuffixTreeApplication, self).__init__(**kwargs)
#         self.suffix_tree_repo = GeneralizedSuffixTreeRepo(self.event_store)
#         self.node_repo = NodeRepo(self.event_store)
#         self.node_child_collection_repo = NodeChildCollectionRepo(self.event_store)
#         self.edge_repo = EdgeRepo(self.event_store)
#         self.stringid_collection_repo = StringidCollectionRepo(self.event_store)
#
#     def close(self):
#         super(SuffixTreeApplication, self).close()
#         self.suffix_tree_repo = None
#         self.node_repo = None
#         self.node_child_collection_repo = None
#         self.edge_repo = None
#         self.stringid_collection_repo = None
#
#     def register_new_suffix_tree(self, case_insensitive=False):
#         """Returns a new suffix tree entity.
#         """
#         suffix_tree = register_new_suffix_tree(case_insensitive=case_insensitive)
#         suffix_tree._node_repo = self.node_repo
#         suffix_tree._node_child_collection_repo = self.node_child_collection_repo
#         suffix_tree._edge_repo = self.edge_repo
#         suffix_tree._stringid_collection_repo = self.stringid_collection_repo
#         return suffix_tree
#
#     def get_suffix_tree(self, suffix_tree_id):
#         """Returns a suffix tree entity, equipped with node and edge repos it (at least at the moment) needs.
#         """
#         suffix_tree = self.suffix_tree_repo[suffix_tree_id]
#         assert isinstance(suffix_tree, GeneralizedSuffixTree)
#         suffix_tree._node_repo = self.node_repo
#         suffix_tree._node_child_collection_repo = self.node_child_collection_repo
#         suffix_tree._edge_repo = self.edge_repo
#         suffix_tree._stringid_collection_repo = self.stringid_collection_repo
#         return suffix_tree
#
#     def find_string_ids(self, substring, suffix_tree_id, limit=None):
#         """Returns a set of IDs for strings that contain the given substring.
#         """
#
#         # Find an edge for the substring.
#         edge, ln = self.find_substring_edge(
#             substring=substring, suffix_tree_id=suffix_tree_id
#         )
#
#         # If there isn't an edge, return an empty set.
#         if edge is None:
#             return set()
#
#         # Get all the string IDs beneath the edge's destination node.
#         string_ids = get_string_ids(
#             node_id=edge.dest_node_id,
#             node_repo=self.node_repo,
#             node_child_collection_repo=self.node_child_collection_repo,
#             stringid_collection_repo=self.stringid_collection_repo,
#             length_until_end=edge.length + 1 - ln,
#             limit=limit,
#         )
#
#         # Return a set of string IDs.
#         return set(string_ids)
#
#     def find_substring_edge(self, substring, suffix_tree_id):
#         """Returns an edge that matches the given substring.
#         """
#         suffix_tree = self.suffix_tree_repo[suffix_tree_id]
#         started = datetime.datetime.now()
#         edge, ln = find_substring_edge(
#             substring=substring, suffix_tree=suffix_tree, edge_repo=self.edge_repo
#         )
#         # if edge is not None:
#         #     print("Got edge for substring '{}': {}".format(substring, edge))
#         # else:
#         #     print("No edge for substring '{}'".format(substring))
#         print(
#             " - searched for edge in {} for substring: '{}'".format(
#                 datetime.datetime.now() - started, substring
#             )
#         )
#         return edge, ln
#
#     def has_substring(self, substring, suffix_tree_id):
#         suffix_tree = self.suffix_tree_repo[suffix_tree_id]
#         return has_substring(
#             substring=substring, suffix_tree=suffix_tree, edge_repo=self.edge_repo
#         )



from typing import Any

from eventsourcing.domain.model.aggregate import BaseAggregateRoot, TAggregate


class Command(BaseAggregateRoot):
    def __init__(self, **kwargs: Any):
        super(Command, self).__init__(**kwargs)
        self._is_done = False

    class Event(BaseAggregateRoot.Event[TAggregate]):
        pass

    class Created(Event[TAggregate], BaseAggregateRoot.Created[TAggregate]):
        pass

    class AttributeChanged(
        Event[TAggregate], BaseAggregateRoot.AttributeChanged[TAggregate]
    ):
        pass

    class Discarded(Event[TAggregate], BaseAggregateRoot.Discarded[TAggregate]):
        pass

    @property
    def is_done(self) -> bool:
        return self._is_done

    def done(self) -> None:
        self.__trigger_event__(self.Done)

    class Done(Event):
        def mutate(self, obj: "Command") -> None:
            obj._is_done = True



import os
from time import sleep, time
from unittest import TestCase
from uuid import uuid4

from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.events import (
    assert_event_handlers_empty,
    clear_event_handlers,
)
from eventsourcing.exceptions import ProgrammingError, RepositoryKeyError
from eventsourcing.system.definition import System
from eventsourcing.system.multiprocess import MultiprocessRunner
from eventsourcing.system.runner import MultiThreadedRunner
from eventsourcing.tests.system_test_fixtures import (
    Examples,
    Order,
    Orders,
    Payment,
    Payments,
    Reservation,
    Reservations,
    create_new_order,
    set_db_uri,
)
from eventsourcing.tests.test_process import ExampleAggregate


class TestSystem(TestCase):
    infrastructure_class = SQLAlchemyApplication

    def test_construct_app(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )
        with system.construct_app(Orders) as app:
            self.assertIsInstance(app, Orders)

        with system.construct_app(Payments) as app:
            self.assertIsInstance(app, Payments)

        with system.construct_app(Reservations) as app:
            self.assertIsInstance(app, Reservations)

    def test_not_infrastructure_class_exception(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=TestCase,
        )
        with self.assertRaises(ProgrammingError):
            system.construct_app(Orders)

    def test_bind(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
        )
        # Check system does not have infrastructure class.
        self.assertIsNone(system.infrastructure_class)

        # Bind system to infrastructure.
        system2 = system.bind(self.infrastructure_class)

        # Check system2 has infrastructure class.
        self.assertIsNotNone(system2.infrastructure_class)

        # Check ProgrammingError on attempt to bind system2.
        with self.assertRaises(ProgrammingError):
            system2.bind(self.infrastructure_class)

    def test_singlethreaded_runner_with_single_application_class(self):
        system = System(
            Orders, setup_tables=True, infrastructure_class=self.infrastructure_class,
        )
        with system as runner:
            orders = runner.get(Orders)
            order_id = orders.create_new_order()
            repository = orders.repository
            self.assertEqual(repository[order_id].id, order_id)

    def test_can_run_if_already_running(self):
        system = System(
            Orders, setup_tables=True, infrastructure_class=self.infrastructure_class,
        )
        with system:
            with self.assertRaises(ProgrammingError):
                with system:
                    pass

        with system as runner:
            with self.assertRaises(ProgrammingError):
                with runner:
                    pass

    def test_multithreaded_runner_with_single_application_class(self):
        system = System(
            Orders, setup_tables=True, infrastructure_class=self.infrastructure_class,
        )
        with MultiThreadedRunner(system) as runner:
            app = runner.get(Orders)
            order_id = app.create_new_order()
            repository = app.repository
            self.assertEqual(repository[order_id].id, order_id)

    def test_multiprocess_runner_with_single_application_class(self):
        system = System(
            Orders, setup_tables=True, infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()
        self.close_connections_before_forking()

        with MultiprocessRunner(system) as runner:
            app = runner.get(Orders)
            order_id = app.create_new_order()
            repository = app.repository
            self.assertEqual(repository[order_id].id, order_id)

    def test_singlethreaded_runner_with_single_pipe(self):
        system = System(
            Orders | Reservations,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )
        with system as runner:
            orders = runner.get(Orders)
            order_id = orders.create_new_order()
            self.assertEqual(orders.repository[order_id].id, order_id)
            reservation_id = Reservation.create_reservation_id(order_id)
            reservations = runner.get(Reservations)
            reservations_repo = reservations.repository
            self.assertEqual(reservations_repo[reservation_id].order_id, order_id)

    def test_multithreaded_runner_with_single_pipe(self):
        system = System(
            Orders | Reservations,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )
        self.set_db_uri()

        # with system as runner:
        with MultiThreadedRunner(system) as runner:
            order_id = create_new_order()
            orders_repo = runner.get(Orders).repository
            self.assertEqual(orders_repo[order_id].id, order_id)
            reservations_repo = runner.get(Reservations).repository
            reservation_id = Reservation.create_reservation_id(order_id)

            patience = 10
            while True:
                try:
                    self.assertEqual(
                        reservations_repo[reservation_id].order_id, order_id
                    )
                except (RepositoryKeyError, AssertionError):
                    if patience:
                        patience -= 1
                        sleep(0.1)
                    else:
                        raise
                else:
                    break

    def test_singlethreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        with system as runner:
            # Create new Order aggregate.
            order_id = create_new_order()

            # Check the order is reserved and paid.
            repository = runner.get(Orders).repository
            assert repository[order_id].is_reserved
            assert repository[order_id].is_paid

    def test_singlethreaded_runner_with_direct_query(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
            use_direct_query_if_available=True,
        )

        with system as runner:
            orders = runner.get(Orders)
            # Create new Order aggregate.
            order_id = create_new_order()

            # Check the order is reserved and paid.
            assert orders.repository[order_id].is_reserved
            assert orders.repository[order_id].is_paid

    def test_multithreaded_runner_with_singleapp_system(self):

        system = System(
            Examples | Examples,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()

        with MultiThreadedRunner(system) as runner:

            app = runner.get(Examples)

            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            assert aggregate.id in app.repository

            # Check the aggregate is moved on.
            retries = 50
            while not app.repository[aggregate.id].is_moved_on:

                sleep(0.1)
                retries -= 1
                assert retries, "Failed to move"

    def test_multithreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()

        with MultiThreadedRunner(system) as runner:

            orders = runner.get(Orders)
            started = time()

            # Create new orders.
            num_orders = 10
            order_ids = []
            for i in range(num_orders):
                order_id = create_new_order()
                order_ids.append(order_id)

            retries = num_orders
            for order_id in order_ids:
                # while not orders.repository[order_id].is_reserved:
                #     sleep(0.1)
                #     retries -= 1
                #     assert retries, "Failed set order.is_reserved"

                while retries and not orders.repository[order_id].is_paid:
                    sleep(0.5)
                    retries -= 1
                    assert retries, "Failed set order.is_paid"

            print(f"Duration: { time() - started :.4f}s")

    def test_clocked_multithreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()

        clock_speed = 10
        with MultiThreadedRunner(system, clock_speed=clock_speed) as runner:

            started = time()

            # Create a new order.
            num_orders = 10
            order_ids = []
            for i in range(num_orders):
                order_id = create_new_order()
                order_ids.append(order_id)
                # sleep(tick_interval / 3)
                # sleep(tick_interval )

            retries = 30 * num_orders
            num_completed = 0
            for order_id in order_ids:
                app = runner.get(Orders)
                while retries and not app.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, (
                        "Failed set order.is_paid (after %s completed)" % num_completed
                    )
                num_completed += 1

        print(f"Duration: { time() - started :.4f}s")

    def test_multiprocessing_singleapp_system(self):

        system = System(
            Examples | Examples,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()

        self.close_connections_before_forking()

        with MultiprocessRunner(system) as runner:

            examples = runner.get(Examples)

            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            assert aggregate.id in examples.repository

            # Check the aggregate is moved on.
            retries = 50
            while not examples.repository[aggregate.id].is_moved_on:

                sleep(0.1)
                retries -= 1
                assert retries, "Failed to move"

    def test_multiprocessing_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        self.set_db_uri()

        with system.construct_app(Orders) as app:
            # Create a new order.
            order_id = create_new_order()
            # Check new order exists in the repository.
            assert order_id in app.repository

        self.close_connections_before_forking()

        with MultiprocessRunner(system):

            with system.construct_app(Orders) as app:
                retries = 50
                while not app.repository[order_id].is_reserved:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved"

                while retries and not app.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid"

    def test_multipipeline_multiprocessing_multiapp(self):

        self.set_db_uri()

        system = System(
            (Orders, Reservations, Orders, Payments, Orders),
            setup_tables=True,
            infrastructure_class=self.infrastructure_class,
        )

        num_pipelines = 2

        pipeline_ids = range(num_pipelines)

        multiprocess_runner = MultiprocessRunner(system, pipeline_ids=pipeline_ids)

        num_orders_per_pipeline = 5
        order_ids = []

        self.close_connections_before_forking()

        # Start multiprocessing system.
        with multiprocess_runner:

            # Create some new orders.
            for _ in range(num_orders_per_pipeline):

                orders = multiprocess_runner.get(Orders)

                for pipeline_id in pipeline_ids:

                    orders.change_pipeline(pipeline_id)

                    order_id = create_new_order()
                    order_ids.append(order_id)

                    sleep(0.05)

            # Wait for orders to be reserved and paid.
            retries = 10 + 10 * num_orders_per_pipeline * len(pipeline_ids)
            for i, order_id in enumerate(order_ids):

                while not orders.repository[order_id].is_reserved:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved {} ({})".format(
                        order_id, i
                    )

                while retries and not orders.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid ({})".format(i)

            # Calculate timings from event timestamps.
            order_aggregates = [orders.repository[oid] for oid in order_ids]
            first_timestamp = min([o.__created_on__ for o in order_aggregates])
            last_timestamp = max([o.__last_modified__ for o in order_aggregates])
            duration = last_timestamp - first_timestamp
            rate = len(order_ids) / float(duration)
            period = 1 / rate
            print(
                "Orders system processed {} orders in {:.3f}s at rate of {:.1f} "
                "orders/s, {:.3f}s each".format(len(order_ids), duration, rate, period)
            )

            # Print min, average, max duration.
            durations = [
                o.__last_modified__ - o.__created_on__ for o in order_aggregates
            ]
            print("Min order processing time: {:.3f}s".format(min(durations)))
            print(
                "Mean order processing time: {:.3f}s".format(
                    sum(durations) / len(durations)
                )
            )
            print("Max order processing time: {:.3f}s".format(max(durations)))

    def set_db_uri(self):
        set_db_uri()

    def close_connections_before_forking(self):
        # Used for closing Django connection before multiprocessing module forks the OS process.
        pass

    def test_payments_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check policy makes payment whenever order is reserved.
        event = Order.Reserved(originator_id=order.id, originator_version=1)

        payment = Payments().policy(repository=fake_repository, event=event)
        assert isinstance(payment, Payment), payment
        assert payment.order_id == order.id

    def test_orders_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check order is not reserved.
        assert not order.is_reserved

        # Reservation created.
        event = Reservation.Created(
            originator_id=uuid4(), originator_topic="", order_id=order.id
        )
        Orders().policy(repository=fake_repository, event=event)

        # Check order is reserved.
        assert order.is_reserved

    def tearDown(self):
        assert_event_handlers_empty()
        clear_event_handlers()
        try:
            del os.environ["DB_URI"]
        except KeyError:
            pass
